import os

import numpy as np

from ..rt_display import DisplayDetect, DisplayString
from ..rt_enum import (
    SDS_EXTENSION_BYTES,
    CategoryMode,
    DisplayDetectModes,
    TypeId,
    TypeRegister,
    gAnsiColors,
)
from ..rt_sds import _build_schema, decompress_dataset_internal

BOX_LIGHT = {
    "UP_AND_RIGHT": "\u2514",
    "HORIZONTAL": "\u2500",
    "VERTICAL": "\u2502",
    "VERTICAL_AND_RIGHT": "\u251C",
}  #: Unicode box-drawing glyphs, light style


class KeyArgsConstructor(object):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


class DictTraversal(KeyArgsConstructor):
    """Traversal suitable for a dictionary. Keys are tree labels, all values
    must be dictionaries as well."""

    def get_children(self, node):
        return list(node[1].items())

    def get_root(self, tree):
        return list(tree.items())[0]

    def get_text(self, node):
        return node[0]


class AttributeTraversal(KeyArgsConstructor):
    """Attribute traversal.

    Uses an attribute of a node as its list of children.
    """

    attribute = "children"  #: Attribute to use.

    def get_children(self, node):
        return getattr(node, self.attribute)


class Style(KeyArgsConstructor):
    """Rendering style for trees."""

    label_format = "{}"  #: Format for labels.

    def node_label(self, text):
        """Render a node text into a label."""
        return self.label_format.format(text)

    def child_head(self, label):
        """Render a node label into final output."""
        return label

    def child_tail(self, line):
        """Render a node line that is not a label into final output."""
        return line

    def last_child_head(self, label):
        """Like :func:`~asciitree.drawing.Style.child_head` but only called
        for the last child."""
        return label

    def last_child_tail(self, line):
        """Like :func:`~asciitree.drawing.Style.child_tail` but only called
        for the last child."""
        return line


class BoxStyle(Style):
    """A rendering style that uses box draw characters and a common layout."""

    gfx = BOX_LIGHT  #: Glyhps to use.
    label_space = 1  #: Space between glyphs and label.
    horiz_len = 4  #: Length of horizontal lines
    indent = 1  #: Indent for subtrees

    def child_head(self, label):
        return (
            " " * self.indent
            + self.gfx["VERTICAL_AND_RIGHT"]
            + self.gfx["HORIZONTAL"] * self.horiz_len
            + " " * self.label_space
            + label
        )

    def child_tail(self, line):
        return " " * self.indent + self.gfx["VERTICAL"] + " " * self.horiz_len + line

    def last_child_head(self, label):
        return (
            " " * self.indent
            + self.gfx["UP_AND_RIGHT"]
            + self.gfx["HORIZONTAL"] * self.horiz_len
            + " " * self.label_space
            + label
        )

    def last_child_tail(self, line):
        return " " * self.indent + " " * len(self.gfx["VERTICAL"]) + " " * self.horiz_len + line


class LeftAligned(KeyArgsConstructor):
    """Creates a renderer for a left-aligned tree.

    Any attributes of the resulting class instances can be set using
    constructor arguments."""

    draw = BoxStyle()
    "The draw style used. See :class:`~asciitree.drawing.Style`."
    traverse = DictTraversal()
    "Traversal method. See :class:`~asciitree.traversal.Traversal`."

    def render(self, node):
        """Renders a node. This function is used internally, as it returns
        a list of lines. Use :func:`~asciitree.LeftAligned.__call__` instead.
        """
        lines = []

        children = self.traverse.get_children(node)
        lines.append(self.draw.node_label(self.traverse.get_text(node)))

        for n, child in enumerate(children):
            child_tree = self.render(child)

            if n == len(children) - 1:
                # last child does not get the line drawn
                lines.append(self.draw.last_child_head(child_tree.pop(0)))
                lines.extend(self.draw.last_child_tail(l) for l in child_tree)
            else:
                lines.append(self.draw.child_head(child_tree.pop(0)))
                lines.extend(self.draw.child_tail(l) for l in child_tree)

        return lines

    def __call__(self, tree):
        """Render the tree into string suitable for console output.

        :param tree: A tree."""
        return "\n".join(self.render(self.traverse.get_root(tree)))


class DisplayNested:

    inline_svg = {
        "array": '<svg xmlns="http://www.w3.org/2000/svg" width="25" height="25" viewBox="0 0 50 50" enable-background="new 0 0 50 50"><path d="M 7 2 L 7 3 L 7 47 L 7 48 L 8 48 L 42 48 L 43 48 L 43 47 L 43 15 L 43 14.59375 L 42.71875 14.28125 L 30.71875 2.28125 L 30.40625 2 L 30 2 L 8 2 L 7 2 z M 9 4 L 29 4 L 29 15 L 29 16 L 30 16 L 41 16 L 41 46 L 9 46 L 9 4 z M 31 5.4375 L 39.5625 14 L 31 14 L 31 5.4375 z"/></svg>',
        "object": '<svg xmlns="http://www.w3.org/2000/svg" width="25" height="25" viewBox="2 2 28 28"><path d="M 16 4 C 9.3844276 4 4 9.3844276 4 16 C 4 22.615572 9.3844276 28 16 28 C 22.615572 28 28 22.615572 28 16 C 28 9.3844276 22.615572 4 16 4 z M 16 6 C 21.534692 6 26 10.465308 26 16 C 26 21.534692 21.534692 26 16 26 C 10.465308 26 6 21.534692 6 16 C 6 10.465308 10.465308 6 16 6 z M 22 16 C 22 19.325562 19.325562 22 16 22 L 16 24 C 20.406438 24 24 20.406438 24 16 L 22 16 z"/></svg>',
        "container": '<svg xmlns="http://www.w3.org/2000/svg" width="25" height="25" viewBox="0 0 50 50"><path d="M 25 0 C 19.3545 0 14.230156 1.0122381 10.4375 2.71875 C 8.5411719 3.5720059 6.9801672 4.6031002 5.84375 5.8125 C 4.7073328 7.0218998 4 8.4567299 4 10 L 4 19.8125 A 1.0001 1.0001 0 0 0 4 20 L 4 29.8125 A 1.0001 1.0001 0 0 0 4 30 L 4 39 C 4 40.54327 4.7073328 41.9781 5.84375 43.1875 C 6.9801672 44.3969 8.5411719 45.427994 10.4375 46.28125 C 14.230156 47.987762 19.3545 49 25 49 C 30.6455 49 35.769844 47.987762 39.5625 46.28125 C 41.458828 45.427994 43.019833 44.3969 44.15625 43.1875 C 45.292667 41.9781 46 40.54327 46 39 L 46 20 L 46 19.90625 L 46 10 C 46 8.4567299 45.292667 7.0218998 44.15625 5.8125 C 43.019833 4.6031002 41.458828 3.5720059 39.5625 2.71875 C 35.769844 1.0122381 30.6455 0 25 0 z M 25 2 C 30.4025 2 35.273406 3.0122619 38.71875 4.5625 C 40.441422 5.3376191 41.800167 6.2431811 42.6875 7.1875 C 43.574833 8.1318189 44 9.0567701 44 10 C 44 10.94323 43.574833 11.868181 42.6875 12.8125 C 41.800167 13.756819 40.441422 14.662381 38.71875 15.4375 C 35.273406 16.987738 30.4025 18 25 18 C 19.5975 18 14.726594 16.987738 11.28125 15.4375 C 9.5585781 14.662381 8.1998328 13.756819 7.3125 12.8125 C 6.4251672 11.868181 6 10.94323 6 10 C 6 9.0567701 6.4251672 8.1318189 7.3125 7.1875 C 8.1998328 6.2431811 9.5585781 5.3376191 11.28125 4.5625 C 14.726594 3.0122619 19.5975 2 25 2 z M 6 14.34375 C 7.1223185 15.486391 8.6228131 16.464729 10.4375 17.28125 C 14.230156 18.987762 19.3545 20 25 20 C 30.6455 20 35.769844 18.987762 39.5625 17.28125 C 41.377187 16.464729 42.877682 15.486391 44 14.34375 L 44 19.84375 A 1.0001 1.0001 0 0 0 44 19.9375 L 44 20 C 44 20.94323 43.574833 21.868181 42.6875 22.8125 C 41.800167 23.756819 40.441422 24.662381 38.71875 25.4375 C 35.273406 26.987738 30.4025 28 25 28 C 19.5975 28 14.726594 26.987738 11.28125 25.4375 C 9.5585781 24.662381 8.1998328 23.756819 7.3125 22.8125 C 6.4251672 21.868181 6 20.94323 6 20 A 1.0001 1.0001 0 0 0 6 19.8125 L 6 14.34375 z M 6 24.34375 C 7.1223185 25.486391 8.6228131 26.464729 10.4375 27.28125 C 14.230156 28.987762 19.3545 30 25 30 C 30.6455 30 35.769844 28.987762 39.5625 27.28125 C 41.377187 26.464729 42.877682 25.486391 44 24.34375 L 44 29.84375 A 1.0001 1.0001 0 0 0 44 29.9375 L 44 30 C 44 30.94323 43.574833 31.868181 42.6875 32.8125 C 41.800167 33.756819 40.441422 34.662381 38.71875 35.4375 C 35.273406 36.987738 30.4025 38 25 38 C 19.5975 38 14.726594 36.987738 11.28125 35.4375 C 9.5585781 34.662381 8.1998328 33.756819 7.3125 32.8125 C 6.4251672 31.868181 6 30.94323 6 30 A 1.0001 1.0001 0 0 0 6 29.8125 L 6 24.34375 z M 6 34.34375 C 7.1223185 35.486391 8.6228131 36.464729 10.4375 37.28125 C 14.230156 38.987762 19.3545 40 25 40 C 30.6455 40 35.769844 38.987762 39.5625 37.28125 C 41.377187 36.464729 42.877682 35.486391 44 34.34375 L 44 39 C 44 39.94323 43.574833 40.868181 42.6875 41.8125 C 41.800167 42.756819 40.441422 43.662381 38.71875 44.4375 C 35.273406 45.987738 30.4025 47 25 47 C 19.5975 47 14.726594 45.987738 11.28125 44.4375 C 9.5585781 43.662381 8.1998328 42.756819 7.3125 41.8125 C 6.4251672 40.868181 6 39.94323 6 39 L 6 34.34375 z"/></svg>',
    }

    def __init__(self):
        self._fmtstart = None
        self._fmtend = None

    # ---------------------------------------------------------------
    # TODO: borrow from global color pallette - probably needs to move out of display
    @property
    def fmtstart(self):
        if self._fmtstart is None:
            self._fmtstart = "" if DisplayDetect.Mode == DisplayDetectModes.Console else gAnsiColors["LightCyan"]
        return self._fmtstart

    # ---------------------------------------------------------------
    @property
    def fmtend(self):
        if self._fmtend is None:
            self._fmtend = "" if DisplayDetect.Mode == DisplayDetectModes.Console else gAnsiColors["Normal"]
        return self._fmtend

    # ---------------------------------------------------------------
    def _map_asciitree(self, data, structure, name=None, info=False):
        structure[self.fmtstart + name + self.fmtend] = {}
        structure = structure[self.fmtstart + name + self.fmtend]

        if not len(data):
            return

        maxlen = len(max(list(data.keys()), key=len))

        for k, v in data.items():

            info_indent = " " * (maxlen - len(k) + 1)

            if hasattr(v, "items"):
                name = f"{k}" if info else f"{k} ({type(v).__name__})"
                self._map_asciitree(v, structure, name=name, info=info)

            else:
                item_formatter = _default_info
                if type(v) == TypeRegister.FastArray or type(v) == np.ndarray:
                    item_formatter = _arr_info

                elif type(v) == TypeRegister.Categorical:
                    item_formatter = _cat_info

                # TODO: Use np.isscalar() here instead?
                elif isinstance(
                    v, (int, float, bool, np.bool_, np.integer, np.floating, str, np.str_, bytes, np.bytes_)
                ):
                    item_formatter = _scalar_info

                item_name = k + f"{info_indent}{item_formatter(v)}"

                structure[item_name] = {}

    def _map_full_paths(self, data, structure, name=None, prefix=""):
        pass
        # structure[prefix+self.fmtstart+name+self.fmtend] = {}
        # structure = structure[prefix+self.fmtstart+name+self.fmtend]

        # maxlen = len(max(list(data.keys()), key=len))

        # for k, v in data.items():

        #    info_indent = " "*(maxlen - len(k)+1)

        #    if hasattr(v, 'items'):
        #        prefix = prefix+name+"."
        #        self._map_full_paths(data, structure, name=f"{k}", prefix=prefix)

        #    else:
        #        structure[prefix+k] = {}

    # ---------------------------------------------------------------
    def _map_htmltree(self, data, structure, name=None, html_str=[], showicon=True):
        structure[name] = {}
        structure = structure[name]

        menu_str = "<li class='menu-header'><a>"
        if showicon:
            menu_str += self.inline_svg["container"]
            # menu_str += "<img src='dbicon.png' />&nbsp;&nbsp;"
        menu_str += f"{name} ({type(data).__name__})</a></li><ul class='sfw-ul'>"
        html_str.append(menu_str)

        for k, v in data.items():
            t = type(v)

            if hasattr(v, "items"):
                self._map_htmltree(v, structure, name=k, html_str=html_str, showicon=showicon)

                html_str.append("</ul></li>")
            else:
                file_str = "<li>"

                if type(v) == TypeRegister.FastArray or type(v) == np.ndarray:
                    if showicon:
                        file_str += self.inline_svg["array"]
                    file_str += f"<p>{k}"
                    file_str += f" {type(v).__name__} {_arr_info(v)}"  # {v.dtype}"

                # TODO: Use np.isscalar() here instead?
                elif isinstance(
                    v, (int, float, bool, np.bool_, np.integer, np.floating, str, np.str_, bytes, np.bytes_)
                ):
                    if showicon:
                        file_str += self.inline_svg["object"]
                    file_str += f"<p>{k}"
                    file_str += f" {v}"
                else:
                    if showicon:
                        file_str += self.inline_svg["object"]
                    file_str += f"<p>{k} {type(v).__name__}"

                file_str += "</p></li>"
                html_str.append(file_str)
                structure[k] = {}

    # ---------------------------------------------------------------
    def _build_nested_html(self, data, name=None):
        html_str = []
        # STYLE
        html_str.append("<html><head><style>")
        html_str.append(
            "a{ padding-top: 15px; line-height: 20px; vertical-align: top; font-weight: bold; cursor: pointer;}"
        )
        html_str.append(".hidelist{ display: none;}")
        html_str.append("li{ margin: 3px auto auto auto; }")
        html_str.append("li.menu-header{ margin: 10px auto auto auto;}")
        html_str.append(
            "li p{ line-height:25px; vertical-align:top; display:inline-block; margin: 0 0 0 5px !important;}"
        )
        html_str.append("svg{display:inline-block;}")
        html_str.append("ul.rt-ul{ list-style-type: none !important; list-style: none !important;}")
        # html_str.append("ul.rt-ul img{ float:left; }")
        # html_str.append("li.listitem-wrap{ width: 100%; height: 20px; line-height: 20px; }")
        html_str.append("</style>")

        # JAVASCRIPT
        html_str.append("<script src='https://ajax.googleapis.com/ajax/libs/jquery/3.3.1/jquery.min.js'></script>")
        html_str.append(
            "<script>$(document).ready(function(){$('a').click(function(){$(this).parent().next().toggleClass('hidelist');return false;});});"
        )
        html_str.append("</script></head><body>")

        if name is None:
            name = type(data).__name__
        # NESTED LISTS
        structure = {}
        html_str.append("<ul class='sfw-ul'>")
        self._map_htmltree(data, structure, name=name, html_str=html_str)
        html_str.append("</ul></body></html>")
        html_str = "".join(html_str)

        return DisplayString(html_str)

    # ---------------------------------------------------------------
    def _build_nested_ascii(self, data, name=None, showpaths=False, info=False):
        if name is None:
            name = type(data).__name__

        structure = {}
        if showpaths:
            self._map_full_paths(data, structure, name=name)
        else:
            self._map_asciitree(data, structure, name=name, info=info)

        ascii_string = LeftAligned()
        return DisplayString(ascii_string(structure))

    def build_nested_html(self, data={}, name=None):
        return self._build_nested_html(data=data, name=name)

    def build_nested_string(self, data={}, name=None, showpaths=False, info=False):
        return self._build_nested_ascii(data=data, name=name, showpaths=showpaths, info=info)


# ---------------------------------------------------------------
def _default_info(item):
    pass


def _scalar_info(item):
    return item


def _arr_info(arr):
    dt = arr.dtype
    typename = dt.name
    sh = str(arr.shape)
    itemsize = str(arr.itemsize)
    return " ".join([typename, sh, itemsize])


def _cat_info(arr):
    if arr.category_mode == CategoryMode.StringArray:
        str_type = arr.category_array.dtype.char
    else:
        str_type = ""

    item_name = f"Categorical {arr.dtype} {arr.shape} {arr.category_mode.name} ({str_type})"
    return item_name


# ---------------------------------------------------------------
def treedir(path, name=None):
    # should we use the struct routine to do this, or just the file names?
    # return TypeRegister.Struct.sload(path, info=True)

    if isinstance(path, bytes):
        path = path.decode()

    if not os.path.isdir(path):
        raise OSError(f"{path} doesn't exist.")

    dirlist = os.listdir(path)
    schema = _build_schema(path, dirlist)

    if name is not None:
        if name in schema:
            schema = schema[name]
        else:
            raise ValueError(f"Could not find {name} in directory {path}")

    if "_root" in schema:
        del schema["_root"]
    dirname = os.path.basename(os.path.normpath(path))

    full_schema = {dirname: schema}
    ascii_string = LeftAligned()
    return DisplayString(ascii_string(full_schema))


# ---------------------------------------------------------------
def _mask_flags(flagnum):
    flagdict = {
        "C_CONTIGUOUS": 0x0001,
        "F_CONTIGUOUS": 0x0002,
        "OWNDATA": 0x0004,
        "FORCECAST": 0x0010,
        "ENSURECOPY": 0x0020,
        "ENSUREARRAY": 0x0040,
        "ELEMENTSTRIDES": 0x0080,
        "ALIGNED": 0x0100,
        "NOTSWAPPED": 0x0200,
        "WRITEABLE": 0x0400,
        "UPDATEIFCOPY": 0x1000,
    }
    for flagname, bitcode in flagdict.items():
        flagdict[flagname] = (flagnum & bitcode) > 0

    return flagdict



#!/usr/bin/env python

__all__ = [
    "get_terminal_size",
]

import os
import platform
import shlex
import struct
import subprocess


def get_terminal_size():
    """getTerminalSize()
    - get width and height of console
    - works on linux,os x,windows,cygwin(windows)
    originally retrieved from:
    http://stackoverflow.com/questions/566746/how-to-get-console-window-width-in-python

    Returns: (x,y) in screen size
    Returns: (None, None) on failure
    """
    current_os = platform.system()
    tuple_xy = None
    if current_os == "Windows":
        tuple_xy = _get_terminal_size_windows()
        if tuple_xy is None:
            tuple_xy = _get_terminal_size_tput()
            # needed for window's python in cygwin's xterm!
    if current_os in ["Linux", "Darwin"] or current_os.startswith("CYGWIN"):
        tuple_xy = _get_terminal_size_linux()
    if tuple_xy is None:
        # we cannot detect the mode, this is likely spyder
        return (None, None)
    return tuple_xy


def _get_terminal_size_windows():
    try:
        from ctypes import create_string_buffer, windll

        # stdin handle is -10
        # stdout handle is -11
        # stderr handle is -12
        h = windll.kernel32.GetStdHandle(-12)
        csbi = create_string_buffer(22)
        res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
        if res:
            (
                bufx,
                bufy,
                curx,
                cury,
                wattr,
                left,
                top,
                right,
                bottom,
                maxx,
                maxy,
            ) = struct.unpack("hhhhHhhhhhh", csbi.raw)
            sizex = right - left + 1
            sizey = bottom - top + 1
            return sizex, sizey
    except:
        pass


def _get_terminal_size_tput():
    # get terminal width
    # src: http://stackoverflow.com/questions/263890/how-do-i-find-the-width-height-of-a-terminal-window
    try:
        cols = int(subprocess.check_call(shlex.split("tput cols")))
        rows = int(subprocess.check_call(shlex.split("tput lines")))
        return cols, rows
    except:
        pass


def _get_terminal_size_linux():
    def ioctl_GWINSZ(fd):
        try:
            import fcntl
            import termios

            cr = struct.unpack("hh", fcntl.ioctl(fd, termios.TIOCGWINSZ, "1234"))
            return cr
        except:
            pass

    cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
    if not cr:
        try:
            fd = os.open(os.ctermid(), os.O_RDONLY)
            cr = ioctl_GWINSZ(fd)
            os.close(fd)
        except:
            pass
    if not cr:
        try:
            cr = (os.environ["LINES"], os.environ["COLUMNS"])
        except:
            return None
    return int(cr[1]), int(cr[0])


if __name__ == "__main__":
    sizex, sizey = get_terminal_size()
    print("width =", sizex, "height =", sizey)



import unittest

from riptable import *
from riptable.rt_datetime import NANOS_PER_HOUR


class DateTimeMath_Test(unittest.TestCase):
    def test_add_scalar(self):
        dtn = DateTimeNano(
            [
                "Inv",
                "19810421 09:06:41.220080800",
                "19980125 03:02:11.912821440",
                "Inv",
                "Inv",
            ],
            from_tz="NYC",
        )
        result = dtn + 1
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)

        result = dtn + 1.0
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)

        result = dtn + [1]
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)

        result = dtn + [1.0]
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)

    def test_iadd_scalar(self):
        dtn = DateTimeNano(
            [
                "Inv",
                "19810421 09:06:41.220080800",
                "19980125 03:02:11.912821440",
                "Inv",
                "Inv",
            ],
            from_tz="NYC",
        )
        dtn += 1
        result = dtn
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)

        dtn = DateTimeNano(
            [
                "Inv",
                "19810421 09:06:41.220080800",
                "19980125 03:02:11.912821440",
                "Inv",
                "Inv",
            ],
            from_tz="NYC",
        )
        dtn += 1.0
        result = dtn
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)

        dtn = DateTimeNano(
            [
                "Inv",
                "19810421 09:06:41.220080800",
                "19980125 03:02:11.912821440",
                "Inv",
                "Inv",
            ],
            from_tz="NYC",
        )
        dtn += [1]
        result = dtn
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)

        dtn = DateTimeNano(
            [
                "Inv",
                "19810421 09:06:41.220080800",
                "19980125 03:02:11.912821440",
                "Inv",
                "Inv",
            ],
            from_tz="NYC",
        )
        dtn += [1.0]
        result = dtn
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)

    def test_add_array(self):
        dtn = DateTimeNano(
            [
                "19881114 22:58:18.875374947",
                "Inv",
                "19781014 09:09:50.558413019",
                "Inv",
                "20180606 14:45:20.799048559",
            ],
            from_tz="NYC",
        )
        result = dtn + full(5, 500)
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[1], 0)

        val = full(5, 500, dtype=np.float32)
        result = dtn + val
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[1], 0)
        val[0] = np.nan
        result = dtn + val
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)
        self.assertEqual(result._fa[1], 0)

    def test_iadd_array(self):
        dtn = DateTimeNano(
            [
                "19881114 22:58:18.875374947",
                "Inv",
                "19781014 09:09:50.558413019",
                "Inv",
                "20180606 14:45:20.799048559",
            ],
            from_tz="NYC",
        )
        dtn += full(5, 500)
        result = dtn
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[1], 0)

        dtn = DateTimeNano(
            [
                "19881114 22:58:18.875374947",
                "Inv",
                "19781014 09:09:50.558413019",
                "Inv",
                "20180606 14:45:20.799048559",
            ],
            from_tz="NYC",
        )
        val = full(5, 500, dtype=np.float32)
        dtn += val
        result = dtn

        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[1], 0)
        val[0] = np.nan
        dtn = DateTimeNano(
            [
                "19881114 22:58:18.875374947",
                "Inv",
                "19781014 09:09:50.558413019",
                "Inv",
                "20180606 14:45:20.799048559",
            ],
            from_tz="NYC",
        )
        dtn += val
        result = dtn
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)
        self.assertEqual(result._fa[1], 0)

    def test_add_dtclass(self):
        dtn = DateTimeNano(
            [
                "20110719 06:30:28.899581754",
                "19800121 20:54:42.763693173",
                "Inv",
                "Inv",
                "19820625 05:59:45.813906699",
            ],
            from_tz="NYC",
        )
        with self.assertRaises(TypeError):
            dtn2 = dtn + dtn

        with self.assertRaises(TypeError):
            dtn2 = dtn + Date(1)

        result = dtn + TimeSpan(1)
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[2], 0)

        result = dtn + DateSpan(1)
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[2], 0)

    def test_sub_dtclass(self):
        # INT64 - INT64 => FLOAT64 (both checked, nan fill)
        dtn = DateTimeNano(
            [
                "Inv",
                "Inv",
                "20010911 05:25:55.782094544",
                "20150217 21:44:37.149415405",
                "Inv",
            ],
            from_tz="NYC",
        )
        val = DateTimeNano(
            [
                "Inv",
                "Inv",
                "20010911 04:25:55.782094544",
                "20150217 20:44:37.149415405",
                "Inv",
            ],
            from_tz="NYC",
        )
        result = dtn - val
        self.assertTrue(isinstance(result, TimeSpan))
        self.assertTrue(isnan(result._fa[0]))

        # INT64 - INT64 => FLOAT64 (both checked for 0, nan fill)
        val = Date(1)
        result = dtn - val
        self.assertTrue(isinstance(result, TimeSpan))
        self.assertTrue(isnan(result._fa[0]))

        # INT64 - INT64 => INT64 (left checked for 0, right checked for INT64 sentinel, 0 fill)
        val = DateSpan(1)
        result = dtn - val
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)

        # INT64 - INT64 (upcast to int) => INT64 (left checked for 0, right checked for INT64 sentinel, 0 fill)
        val = TimeSpan(1)
        result = dtn - val
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)

    def test_sub_scalar(self):
        dtn = DateTimeNano(
            [
                "Inv",
                "19810421 09:06:41.220080800",
                "19980125 03:02:11.912821440",
                "Inv",
                "Inv",
            ],
            from_tz="NYC",
        )
        result = dtn - 1
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)
        self.assertEqual(dtn._fa[1] - 1, result._fa[1])

        result = dtn - 1.0
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)
        self.assertEqual(dtn._fa[1] - 1, result._fa[1])

        result = dtn - [1]
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)
        self.assertEqual(dtn._fa[1] - 1, result._fa[1])

        result = dtn - [1.0]
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)
        self.assertEqual(dtn._fa[1] - 1, result._fa[1])

    def test_isub_scalar(self):
        dtn = DateTimeNano(
            [
                "Inv",
                "19810421 09:06:41.220080800",
                "19980125 03:02:11.912821440",
                "Inv",
                "Inv",
            ],
            from_tz="NYC",
        )
        dtn -= 1
        result = dtn
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)

        dtn = DateTimeNano(
            [
                "Inv",
                "19810421 09:06:41.220080800",
                "19980125 03:02:11.912821440",
                "Inv",
                "Inv",
            ],
            from_tz="NYC",
        )
        dtn -= 1.0
        result = dtn
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)

        dtn = DateTimeNano(
            [
                "Inv",
                "19810421 09:06:41.220080800",
                "19980125 03:02:11.912821440",
                "Inv",
                "Inv",
            ],
            from_tz="NYC",
        )
        dtn -= [1]
        result = dtn
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)

        dtn = DateTimeNano(
            [
                "Inv",
                "19810421 09:06:41.220080800",
                "19980125 03:02:11.912821440",
                "Inv",
                "Inv",
            ],
            from_tz="NYC",
        )
        dtn -= [1.0]
        result = dtn
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)

    def test_sub_array(self):
        dtn = DateTimeNano(
            [
                "19881114 22:58:18.875374947",
                "Inv",
                "19781014 09:09:50.558413019",
                "Inv",
                "20180606 14:45:20.799048559",
            ],
            from_tz="NYC",
        )
        result = dtn - full(5, 500)
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[1], 0)

        val = full(5, 500, dtype=np.float32)
        result = dtn - val
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[1], 0)
        val[0] = np.nan
        result = dtn - val
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)
        self.assertEqual(result._fa[1], 0)

    def test_isub_array(self):
        dtn = DateTimeNano(
            [
                "19881114 22:58:18.875374947",
                "Inv",
                "19781014 09:09:50.558413019",
                "Inv",
                "20180606 14:45:20.799048559",
            ],
            from_tz="NYC",
        )
        dtn -= full(5, 500)
        result = dtn
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[1], 0)

        dtn = DateTimeNano(
            [
                "19881114 22:58:18.875374947",
                "Inv",
                "19781014 09:09:50.558413019",
                "Inv",
                "20180606 14:45:20.799048559",
            ],
            from_tz="NYC",
        )
        val = full(5, 500, dtype=np.float32)
        dtn -= val
        result = dtn
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[1], 0)
        val[0] = np.nan
        dtn = DateTimeNano(
            [
                "19881114 22:58:18.875374947",
                "Inv",
                "19781014 09:09:50.558413019",
                "Inv",
                "20180606 14:45:20.799048559",
            ],
            from_tz="NYC",
        )
        dtn -= val
        result = dtn
        self.assertTrue(isinstance(result, DateTimeNano))
        self.assertEqual(result._fa[0], 0)
        self.assertEqual(result._fa[1], 0)

    def test_timespan_add_scalar(self):
        ts = TimeSpan([10, 300, 456, 20, np.nan], unit="m")
        result = ts + 1
        self.assertTrue(isinstance(result, TimeSpan))
        self.assertTrue(isnan(result._fa[4]))

        result = ts + [1]
        self.assertTrue(isinstance(result, TimeSpan))
        self.assertTrue(isnan(result._fa[4]))

        result = ts + FA(1)
        self.assertTrue(isinstance(result, TimeSpan))
        self.assertTrue(isnan(result._fa[4]))

        result = ts + 1.0
        self.assertTrue(isinstance(result, TimeSpan))
        self.assertTrue(isnan(result._fa[4]))

        result = ts + [1.0]
        self.assertTrue(isinstance(result, TimeSpan))
        self.assertTrue(isnan(result._fa[4]))

        result = ts + FA(1.0)
        self.assertTrue(isinstance(result, TimeSpan))
        self.assertTrue(isnan(result._fa[4]))

    def test_timespan_add_array(self):
        ts = TimeSpan([10, 300, 456, 20, np.nan], unit="m")
        arr = full(5, 1_000_000_000, dtype=np.float64)
        result = ts + arr
        self.assertTrue(isinstance(result, TimeSpan))
        self.assertTrue(isnan(result._fa[4]))
        arr[0] = np.nan
        result = ts + arr
        self.assertTrue(isinstance(result, TimeSpan))
        self.assertTrue(isnan(result._fa[4]))
        self.assertTrue(isnan(result._fa[0]))
        arr = arr.astype(np.int64)
        result = ts + arr
        self.assertTrue(isinstance(result, TimeSpan))
        self.assertTrue(isnan(result._fa[4]))
        self.assertTrue(isnan(result._fa[0]))

    def test_dtn_compare_date(self):

        dtn = DateTimeNano(["20190227"], from_tz="NYC", to_tz="NYC").repeat(24)
        dtn = dtn + TimeSpan(arange(24), unit="h")
        self.assertTrue(bool(np.all(dtn.hour == arange(24))))

        LT = dtn < Date("20190228")
        self.assertTrue(bool(np.all(LT)))

        GT = dtn > Date("20190227")
        self.assertTrue(bool(np.all(GT[1:])))
        self.assertFalse(GT[0])

        dtn = dtn + NANOS_PER_HOUR
        EQ = dtn == Date("20190228")
        self.assertTrue(EQ[-1])

        dtn = DateTimeNano(["20190227"], from_tz="GMT", to_tz="GMT").repeat(24)
        dtn = dtn + TimeSpan(arange(24), unit="h")
        self.assertTrue(bool(np.all(dtn.hour == arange(24))))

        LT = dtn < Date("20190228")
        self.assertTrue(bool(np.all(LT)))

        GT = dtn > Date("20190227")
        self.assertTrue(bool(np.all(GT[1:])))
        self.assertFalse(GT[0])

        dtn = dtn + NANOS_PER_HOUR
        EQ = dtn == Date("20190228")
        self.assertTrue(EQ[-1])

    def test_diff(self):
        dt = Date(["201910%02d" % i for i in range(1, 31)])
        d = dt.diff()
        self.assertTrue(isinstance(d, DateSpan))
        dt = utcnow(30)
        d = dt.diff()
        self.assertTrue(isinstance(d, TimeSpan))


if __name__ == "__main__":
    tester = unittest.main()



#! /usr/bin/env python

import os
import sys
from contextlib import contextmanager
from enum import IntEnum

__all__ = [
    "LikertDecision",
    "redirectStderr2StdoutCtx",
    "redirectStdoutCtx",
    "redirectStdoutAndErrorCtx",
    "chdirCtx",
    "new_array_function",
]


# TODO refactor LikerDecision to ISO3166CountryCodes
# Instead of Likert scale, which can be confusing when considering categorical sort order,
# use countries from ISO-3166 country codes as a replacement
class LikertDecision(IntEnum):
    """A Likert scale with the typical five-level Likert item format."""

    StronglyAgree = 44
    Agree = 133
    Disagree = 75
    StronglyDisagree = 1
    NeitherAgreeNorDisagree = 144


@contextmanager
def redirectStderr2StdoutCtx():
    origStderr = sys.stderr
    sys.stderr = sys.stdout
    try:
        yield
    except:
        sys.stderr = origStderr
        sys.stdout.flush()
        raise
    else:
        sys.stderr = origStderr
        sys.stdout.flush()


@contextmanager
def redirectStdoutCtx(filenameOrStream, mode="w"):
    if filenameOrStream is None:
        yield sys.stdout
        return
    if hasattr(filenameOrStream, "write"):
        f = filenameOrStream
    else:
        f = open(filenameOrStream, mode)
    origStdout = sys.stdout
    sys.stdout = f
    try:
        yield f
    except:
        sys.stdout = origStdout
        if f != filenameOrStream:
            f.close()
        raise
    else:
        sys.stdout = origStdout
        if f != filenameOrStream:
            f.close()


@contextmanager
def redirectStdoutAndErrorCtx(filenameOrStream, mode="w"):
    if filenameOrStream is None:
        yield sys.stdout
        return
    if hasattr(filenameOrStream, "write"):
        f = filenameOrStream
    else:
        f = open(filenameOrStream, mode)
    origStdout = sys.stdout
    origStderr = sys.stderr
    sys.stdout = f
    sys.stderr = f
    try:
        yield f
    except:
        sys.stderr = origStderr
        sys.stdout = origStdout
        if f != filenameOrStream:
            f.close()
        raise
    else:
        sys.stderr = origStderr
        sys.stdout = origStdout
        if f != filenameOrStream:
            f.close()


@contextmanager
def chdirCtx(newdir):
    """
    print os.getcwd()
    with chdirCtx(newdir) as (od, nd):
      print od, nd, os.getcwd()
    print os.getcwd()
    """
    if not os.path.isdir(newdir):
        raise IOError("chdirCtx(%r): not a directory" % newdir)
    olddir = os.getcwd()
    try:
        os.chdir(newdir)
        yield (olddir, newdir)
    except:
        os.chdir(olddir)
        raise
    else:
        os.chdir(olddir)


@contextmanager
def new_array_function(kls: type):
    """A context manager that allows using the new array function code path given a class `kls`
    that has a feature flag ``NEW_ARRAY_FUNCTION_ENABLED`` to enable the NEP 18 array function
    machinery."""
    try:
        kls.NEW_ARRAY_FUNCTION_ENABLED = True
        yield
    finally:
        kls.NEW_ARRAY_FUNCTION_ENABLED = False



import unittest

import pandas as pd

from riptable import Dataset
from riptable.Utils.pandas_utils import dataset_as_pandas_df, dataset_from_pandas_df

# Parse the pandas library version into a tuple of integers.
pd_version = tuple(map(int, pd.__version__.split(".")))

# Well-known pandas version(s) that we may need to recognize
# within the tests below.
pd_ver_0_24 = (0, 24, 0)


class TestPandasUtils(unittest.TestCase):
    def test_dataset_from_pandas_df_warn(self):
        df = pd.DataFrame({"a": [1, 2, 3]})

        if pd_version >= pd_ver_0_24:
            with self.assertWarns(FutureWarning):
                ds = dataset_from_pandas_df(df)
        else:
            ds = dataset_from_pandas_df(df)

    def test_dataset_as_pandas_df_warn(self):
        ds = Dataset({"a": [1, 2, 3]})

        if pd_version >= pd_ver_0_24:
            with self.assertWarns(FutureWarning):
                df = dataset_as_pandas_df(ds)
        else:
            df = dataset_as_pandas_df(ds)


if __name__ == "__main__":
    tester = unittest.main()



"""
Logic / helper functions used throughout the riptable benchmark suite.
"""
import numbers
from typing import List, Mapping, Optional, Tuple

import numpy as np

import riptable as rt

try:
    from functools import cached_property
except ImportError:
    cached_property = property

import weakref
from threading import RLock


# Cached_weakref_property, based on functools.cached_property.
# Stores weakrefs to values in a WeakValueDictionary to avoid cycles.
# Particularly valueable when caching views of self in self, which currently produces an
# unbreakable cycle in NumPy (see https://github.com/numpy/numpy/issues/6581).
class cached_weakref_property:
    _NO_OBJECT = object()
    _WEAKREF_CACHE_NAME = "_weakref_cache"

    def __init__(self, func):
        self.func = func
        self.attrname = None
        self.__doc__ = func.__doc__
        self.lock = RLock()

    def __set_name__(self, owner, name):
        if self.attrname is None:
            self.attrname = name
        elif name != self.attrname:
            raise TypeError(f"Cannot reassign name from {self.attrname} to {name}")

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        if self.attrname is None:
            raise TypeError("Missing call to __set_name__() to set attribute name")
        # First, obtain the weakref cache using double-checked locking patttern.
        cache = getattr(instance, cached_weakref_property._WEAKREF_CACHE_NAME, None)
        if cache is None:
            with self.lock:
                cache = getattr(instance, cached_weakref_property._WEAKREF_CACHE_NAME, None)
                if cache is None:
                    try:
                        cache = weakref.WeakValueDictionary()
                        setattr(instance, cached_weakref_property._WEAKREF_CACHE_NAME, cache)
                    except AttributeError:
                        raise TypeError(
                            f"Cannot create f{cached_weakref_property._WEAKREF_CACHE_NAME} attribute on {type(instance)}"
                        ) from None
        # Second, obtain the unwrapped value, always under lock since WeakValueDictonary is not thread-safe.
        with self.lock:
            val = cache.get(self.attrname, cached_weakref_property._NO_OBJECT)
            if val is cached_weakref_property._NO_OBJECT:
                # Definintely not there so while under lock store cached value
                val = self.func(instance)
                cache[self.attrname] = val
        return val


_SEED = 1234
"""
Value used to seed random number generators so we get "random" data for
the benchmarks while also allowing the benchmarks to be repeatable.
"""

_INT_16_MAX = np.iinfo(np.int16).max
_INT_32_MAX = np.iinfo(np.int32).max
_INT_MAX = np.int64(1 << 33)  # 8_589_934_592


def trial_size(low=250, high=_INT_16_MAX, scale_factor=2):
    """Generates a list of input sizes to benchmark against."""
    lst, i = [low], 0
    while lst[i] <= high:
        i += 1
        lst.append(lst[i - 1] * scale_factor)
    return lst


_DOUBLING_TRIAL_16 = trial_size(high=_INT_16_MAX, scale_factor=2)
_DOUBLING_TRIAL_32 = trial_size(high=_INT_32_MAX, scale_factor=2)
_DOUBLING_TRIAL_MAX = trial_size(high=_INT_MAX, scale_factor=2)


_dtypes_by_group = {k: [np.dtype(x) for x in v] for (k, v) in np.typecodes.items()}

# Add a few additional categories to our dtypes dictionary for convenience.
_dtypes_by_group["Boolean"] = [np.dtype("?")]
_dtypes_by_group["StandardFloat"] = [np.dtype(x) for x in np.typecodes["Float"] if np.dtype(x).itemsize > 2]
_dtypes_by_group["RiptableNumeric"] = _dtypes_by_group["AllInteger"] + _dtypes_by_group["StandardFloat"]

dtypes_by_group: Mapping[str, List[np.dtype]] = _dtypes_by_group
"""
``np.typecodes`` but the value for each entry is a list containing the
dtypes corresponding to the typecode(s) for the original entry.

See Also
--------
riptable.rt_enum.NumpyCharTypes
"""


def zeros_eager(shape, dtype: np.dtype) -> np.ndarray:
    # Allocate a zeroed-out array but use np.full rather than np.zeros
    # as the latter doesn't actually touch any of the data pages during
    # allocation; that leads to page faults occurring during the benchmark.
    return np.full(shape, 0, dtype=dtype)


def integer_valid_range(dtype: np.dtype) -> Tuple[int, int]:
    """
    Given an integer dtype, return the range [lo, hi] of valid values
    representable by that dtype.

    Parameters
    ----------
    dtype : data-type
        An integer dtype.
    """
    if not issubclass(dtype.type, numbers.Integral):
        raise ValueError(f"'{dtype}' is not an integral/integer dtype.")

    # Handle bool specially
    if np.issubdtype(dtype, bool):
        return (0, 1)

    # Get the info for this integer dtype
    dtype_info = np.iinfo(dtype)

    # Get the invalid value for this integer dtype
    try:
        invalid_value = rt.INVALID_DICT[dtype.num]
    except KeyError:
        # This type doesn't have an invalid, so return it's full range.
        return (dtype_info.min, dtype_info.max)

    # The invalid value should be either the min or max value of the integer.
    # Determine which one it is, increment/decrement the integer range and return that.
    if invalid_value == dtype_info.min:
        return (dtype_info.min + 1, dtype_info.max)
    elif invalid_value == dtype_info.max:
        return (dtype_info.min, dtype_info.max - 1)
    else:
        raise ValueError(
            f"Unable to determine the valid range for dtype '{dtype}'. invalid_value={invalid_value}\tiinfo={dtype_info}"
        )


def integer_range(dtype: np.dtype, include_invalid: bool = False) -> Tuple[int, int]:
    """
    Given an integer dtype, return the range [lo, hi] of values representable
    by the type, optionally excluding the invalid value (if any) for the type.
    """
    if not issubclass(dtype.type, numbers.Integral):
        raise ValueError(f"'{dtype}' is not an integral/integer dtype.")

    # Handle bool specially
    if np.issubdtype(dtype, bool):
        return (0, 1)

    # Determine the range based on whether the caller wants to include
    # or exclude the invalid value for this dtype.
    if include_invalid:
        dtype_info = np.iinfo(dtype)
        return (dtype_info.min, dtype_info.max)
    else:
        return integer_valid_range(dtype)


def rand_integers(
    gen: np.random.Generator,
    size: Optional[int] = None,
    dtype: np.dtype = np.int64,
    include_invalid: bool = False,
) -> np.ndarray:
    """
    Generate a random array of integers with the specified length and dtype.

    The elements of the array will span the representable range of the dtype,
    optionally including the 'invalid' value for the type. The elements of the
    array are drawn from the 'discrete uniform' distribution.
    """
    # Determine the range for the dtype.
    lo, hi = integer_range(dtype, include_invalid)
    return gen.integers(lo, hi, size, dtype=dtype, endpoint=True)


def rand_floats(
    gen: np.random.Generator,
    low: float = 0.0,
    high: float = 1.0,
    size: Optional[int] = None,
    dtype: np.dtype = np.float64,
    include_invalid: bool = False,
) -> np.ndarray:
    """
    Generate a random array of floating-point values with the specified length and dtype.

    The elements of the array are drawn from the uniform distribution over the
    range ``[low, high)``.
    """
    # Generate a random array for the given floating-point type.
    arr = gen.uniform(low=low, high=high, size=size).astype(dtype)

    # If we're including invalid values (NaN for floating-point types), draw a random integer
    # indicating how many elements we'll set to NaN; then generate a random integer array of
    # that length whose elements will be a fancy index we'll use to assign NaNs into the generated
    # floating-point array.
    # NOTE: The nancount we generate is approximate because we'll don't enforce that all the
    #       elements of the fancy index are unique.
    if include_invalid:
        nancount = gen.integers(0, size, endpoint=True)
        nan_indices = gen.integers(0, size, size=nancount)
        arr[nan_indices] = np.nan

    return arr


# TODO: Define functions for determining things like array sizes to parameterize benchmarks with.
#       If possible, utilize either the asv API or the general Python API to determine how much memory
#       the current machine has installed and use that to determine the upper bound for these sizes;
#       the exact value will depend on things like the dtype, how many arrays are needed for that particular
#       benchmark, etc.


# TODO: Define functions for generating random inputs (using the numpy.random.Generator API)
#       For testing any set-based functionality (including ismember, Grouping, merge, key-alignment)
#       we'll want a way to generate arrays where the elements' multiplicities are drawn from
#       different distributions (e.g. uniform, zipf, geometric) to more-closely model real-world datasets.



# -*- coding: utf-8 -*-
from typing import NamedTuple


class RunResult(NamedTuple):
    """
    Timing results for a benchmark run (statistics calculated from multiple
    invocations of a function to be benchmarked).
    """

    min_ns: int
    """
    The minimum duration, in nanoseconds, of any single invocation of the function.
    """

    median_ns: int
    """
    The median duration, in nanoseconds, of any single invocation of the function.
    """

    max_ns: int
    """
    The maximum duration, in nanoseconds, of any single invocation of the function.
    """

    est_mean_ns: int
    """
    The estimated population mean, in nanoseconds, of any single invocation of the function.
    """

    est_variance: float
    """
    The estimated population variance, in nanoseconds^2, from the `est_mean_duration`
    for each invocation of the function.
    """



import unittest

from riptable import *
from riptable.rt_enum import GROUPBY_KEY_PREFIX

funcs = [np.sqrt, lambda x: x + 1]
reduce_funcs = [sum, np.min, np.max]


def arr_eq(a, b):
    return bool(np.all(a == b))


def arr_all(a):
    return bool(np.all(a))


class transform_apply_test(unittest.TestCase):
    def test_apply(self):
        ds = Dataset({"col_" + str(i): np.random.rand(30) for i in range(5)})
        ds.gb_col = np.random.choice(["a", "b", "c", "d", "e"], 30)
        ds_sorted = ds.sort_copy("gb_col")
        gb = ds_sorted.gb("gb_col")

        for f in reduce_funcs:
            gb_result = gb.apply(f)
            self.assertTrue(isinstance(gb_result, Dataset))
            for gb_rownum, gb_unique in enumerate(["a", "b", "c", "d", "e"]):
                mask = ds_sorted.gb_col == gb_unique
                for i in range(5):
                    col_name = "col_" + str(i)
                    col_slice = ds_sorted[col_name][mask]
                    ds_item = f(col_slice)
                    gb_item = gb_result[col_name][gb_rownum]
                    self.assertAlmostEqual(ds_item, gb_item, msg=f"func:{f} {ds_item} {gb_item}")

    def test_apply_arrays(self):
        ds = Dataset({"col_" + str(i): np.random.rand(30) for i in range(5)})
        ds.gb_col = np.random.choice(["a", "b", "c", "d", "e"], 30)
        ds_sorted = ds.sort_copy("gb_col")
        gb = ds_sorted.gb("gb_col")

        for f in funcs:
            gb_result = gb.apply(f)
            # self.assertTrue(isinstance(gb_result, list))
            for i in range(5):
                colname = "col_" + str(i)
                acol = f(ds_sorted[colname])
                bcol = gb_result[i]
                self.assertTrue(isinstance(bcol, np.ndarray))
                for idx, item in enumerate(acol):
                    self.assertAlmostEqual(item, bcol[idx], msg=f"{item} {bcol[idx]} {f.__name__}")

    def test_docstring_examples(self):

        ds = Dataset({"A": "a a b".split(), "B": [1, 2, 3], "C": [4, 6, 5]})
        g = GroupBy(ds, "A")

        result = g.apply(lambda x: x.sum())
        self.assertTrue(isinstance(result, Dataset))
        self.assertEqual(result._ncols, 3)
        self.assertTrue(arr_eq(result.A, FA(["a", "b"])))
        self.assertTrue(arr_eq(result.B, FA([3, 3])))
        self.assertTrue(arr_eq(result.C, FA([10, 5])))

        result = g.apply(lambda x: {"B": x.B.sum()})
        self.assertTrue(isinstance(result, Dataset))
        self.assertEqual(result._ncols, 2)
        self.assertTrue(arr_eq(result.A, FA(["a", "b"])))
        self.assertTrue(arr_eq(result.B, FA([3, 3])))

        result = g.apply(lambda x: x.max() - x.min())
        self.assertTrue(isinstance(result, Dataset))
        self.assertEqual(result._ncols, 3)
        self.assertTrue(arr_eq(result.A, FA(["a", "b"])))
        self.assertTrue(arr_eq(result.B, FA([1, 0])))
        self.assertTrue(arr_eq(result.C, FA([2, 0])))

        result = g.apply(lambda x: Dataset({"val": [x.C.max() - x.B.min()]}))
        self.assertTrue(isinstance(result, Dataset))
        self.assertEqual(result._ncols, 2)
        self.assertTrue(arr_eq(result.A, FA(["a", "b"])))
        self.assertTrue(arr_eq(result.val, FA([5, 2])))

        def userfunc(x):
            x.Sub = x.C - x.B
            return x

        result = g.apply(userfunc)
        self.assertTrue(isinstance(result, Dataset))
        self.assertEqual(result._ncols, 4)
        self.assertTrue(arr_eq(ds.A, result.A))
        self.assertTrue(arr_eq(ds.B, result.B))
        self.assertTrue(arr_eq(ds.C, result.C))
        self.assertTrue(arr_eq(result.Sub, FA([3, 4, 2])))

    def test_apply_catcol_filter(self):

        ds = Dataset({"keycol": FA(["a", "a", "b", "b", "a"]), "data": np.random.rand(5)})
        ds.keycol = Categorical(ds.keycol)
        f = FA([False, False, True, True, True])

        result_prefilter = ds.gb("keycol", filter=f).apply(sum)
        result_postfilter = ds.gb("keycol").apply(sum, filter=f)
        self.assertTrue(result_prefilter.equals(result_postfilter))

    def test_apply_userfunc(self):
        symbols = np.random.choice(["AAPL", "AMZN", "NFLX"], 20)
        spot = zeros(20)
        trade_size = np.random.choice([10.0, 25.0, 5.0], 20)
        ds = Dataset({"symbol": symbols, "spot": spot, "trade_size": trade_size})

        aapl_mask = ds.symbol == "AAPL"
        tvals = (np.random.rand(sum(aapl_mask)) * 5) + 180
        ds.spot[aapl_mask] = tvals

        amzn_mask = ds.symbol == "AMZN"
        tvals = (np.random.rand(sum(amzn_mask)) * 15) + 1830
        ds.spot[amzn_mask] = tvals

        nflx_mask = ds.symbol == "NFLX"
        tvals = (np.random.rand(sum(nflx_mask)) * 15) + 1830
        ds.spot[nflx_mask] = tvals

    def test_apply_userfunc_kwargs(self):
        def addfunc(x, addent=0):
            x.data = x.data + 100
            return x

        arr = np.random.choice(["a", "b", "c"], 20)
        c = Categorical(arr)
        data = arange(20)

        ds_norm = Dataset({"keycol": arr, "data": data})
        ds_result = ds_norm.gb("keycol").apply(addfunc, addent=100)

        c = Categorical(arr, ordered=False)
        c_result = c.apply(addfunc, ds_norm.data, addent=100)

        ds_cat = Dataset({"keycol": Categorical(arr, ordered=False), "data": data})
        dsc_result = ds_cat.gb("keycol").apply(addfunc, addent=100)

        self.assertTrue(arr_eq(ds_result.data, c_result.data))
        self.assertTrue(arr_eq(c_result.data, dsc_result.data))
        self.assertTrue(arr_eq(np.sort(c_result.data) - 100, data))

    def test_apply_userfunc_multikey(self):
        st_arr = np.random.choice(["a", "b", "c"], 25)
        int_arr = np.random.choice([10, 20, 30], 25)
        data = arange(25)

        c = Categorical([st_arr, int_arr])

        ds_norm = Dataset(c.expand_dict)
        ds_norm.data = data

        ds_cat = Dataset({"keycol": Categorical([st_arr, int_arr]), "data": data})

        c_result = c.apply(lambda x: x + 100, ds_norm.data)
        ds_result = ds_norm.gb(["key_0", "key_1"]).apply(lambda x: x + 100)
        dsc_result = ds_cat.gb("keycol").apply(lambda x: x + 100)

        self.assertTrue(arr_eq(c_result.data, ds_result.data))
        self.assertTrue(arr_eq(ds_result.data, dsc_result.data))

    def test_apply_filter(self):
        st_arr = np.random.choice(["a", "b", "c"], 25)
        data = arange(25)
        f = logical(data % 3)
        ds = Dataset({"strs": st_arr, "data": data})
        ds_result = ds.gb("strs", lex=True).apply(sum, filter=f)

        c = Categorical(ds.strs)
        c_result = c.apply(sum, ds.data, filter=f)

        self.assertTrue(ds_result.equals(c_result))

    def test_apply_filter_multikey(self):
        st_arr = np.random.choice(["a", "b", "c"], 25).astype("S")
        int_arr = np.random.choice([10, 20, 30], 25)
        data = arange(25)
        f = logical(arange(25) % 3)

        c = Categorical([st_arr, int_arr])

        ds_norm = Dataset(c.expand_dict)
        ds_norm.data = data

        ds_cat = Dataset({GROUPBY_KEY_PREFIX: Categorical([st_arr, int_arr]), "data": data})

        c_result = c.apply(np.sum, data, filter=f)
        ds_result = ds_norm.gbu(["key_0", "key_1"]).apply(np.sum, data, filter=f)
        dsc_result = ds_cat.gbu(GROUPBY_KEY_PREFIX).apply(np.sum, data, filter=f)

        self.assertTrue(c_result.equals(ds_result))
        self.assertTrue(ds_result.equals(dsc_result))

    def test_compare_agg(self):
        st_arr = np.random.choice(["a", "b", "c"], 25)
        data = arange(25)
        data.set_name("data")

        c = Categorical(st_arr)

        agg_result = c.agg({"data": [mean, sum]}, data)

        aggfunc = lambda x: {"Mean": x.data.mean(), "Sum": x.data.sum()}
        app_result = c.apply(aggfunc, data)

        self.assertTrue(arr_eq(agg_result.Mean.data, app_result.Mean))
        self.assertTrue(arr_eq(agg_result.Sum.data, app_result.Sum))

    def test_apply_errors(self):

        st_arr = np.random.choice(["a", "b", "c"], 25)
        c = Categorical(st_arr)

        # no data
        with self.assertRaises(ValueError):
            _ = c.apply(sum)

        # not callable
        with self.assertRaises(TypeError):
            _ = c.apply("a", arange(25))

    def test_appply_reduce_sameid(self):
        data = Dataset({"id": np.array([0, 0, 1, 1]), "val": np.array([0, 10, 20, 30])})

        def func_nr(x, y):
            return x.cumsum() + y

        def func_r(x, y):
            return x.sum() + y[0]

        result = Cat(data.id).apply_nonreduce(func_nr, (data.val, data.id))
        self.assertTrue(arr_eq(result["val"], [0, 10, 21, 51]))

        result = Cat(data.id).apply_reduce(func_r, (data.val, data.id))
        self.assertTrue(arr_eq(result["val"], [10, 51]))

    def test_apply_ema(self):

        correct = FastArray(
            [
                150.63090515,
                -3.11271882,
                0.0,
                -43.95726776,
                -36.96357727,
                -58.97815704,
                -80.67763519,
                33.27441788,
                32.07076263,
                -38.05633926,
                -31.82532883,
                505.58901978,
                -53.8103981,
                42.10805893,
                -289.54257202,
                -433.78964233,
                0.0,
                210.54029846,
                31.44789696,
                105.86235809,
                65.31745148,
                -111.80770111,
                22.91087723,
                288.18487549,
                -2536.56005859,
            ],
            dtype=np.float32,
        )

        sym = FastArray(
            # Top SPX constituents by weight as of 20200414
            [
                b"MSFT",
                b"AAPL",
                b"AMZN",
                b"FB",
                b"JNJ",
                b"GOOG",
                b"GOOGL",
                b"BRK.B",
                b"PG",
                b"JPM",
                b"V",
                b"INTC",
                b"UNH",
                b"VZ",
                b"MA",
                b"T",
                b"HD",
                b"MRK",
                b"PFE",
                b"PEP",
                b"BAC",
                b"DIS",
                b"KO",
                b"WMT",
                b"CSCO",
            ],
            dtype="|S8",
        )
        sym = Categorical(sym)

        org = FastArray(
            [
                0.0,
                150.63090515,
                -3.11271882,
                -80.67763519,
                505.58901978,
                33.27441788,
                32.07076263,
                -43.95726776,
                105.86235809,
                -53.8103981,
                22.91087723,
                -31.82532883,
                -111.80770111,
                288.18487549,
                -289.54257202,
                65.31745148,
                -38.05633926,
                -433.78964233,
                31.44789696,
                210.54029846,
                0.0,
                -58.97815704,
                42.10805893,
                -2536.56005859,
                -36.96357727,
            ]
        )

        time = FastArray(
            [
                3.98803711,
                10.36206055,
                25.64599609,
                26.74609375,
                27.3269043,
                27.71777344,
                27.88891602,
                27.88891602,
                30.08300781,
                30.32299805,
                31.14013672,
                32.32885742,
                32.96582031,
                35.02587891,
                35.02587891,
                35.30810547,
                37.74584961,
                38.34008789,
                38.37207031,
                38.37207031,
                39.13793945,
                40.09399414,
                41.65991211,
                42.22998047,
                47.26000977,
            ]
        )

        ds = Dataset({"sym": sym, "org": org, "time": time})

        def get_ema_userfunc(x, decay_rate=None):
            t = x.time
            v = x.org
            arrlen = len(v)

            ema = empty(arrlen)
            lastEma = 0
            lastTime = 0

            for i in range(arrlen):
                ema[i] = v[i] + lastEma * np.exp(-decay_rate * (t[i] - lastTime))
                lastEma = ema[i]
                lastTime = t[i]

            return {"ema": ema}

        # this is when two inputs specified via a tuple, but no output
        def get_ema_userfunc2(v, t, decay_rate=None):
            arrlen = len(v)
            ema = empty(arrlen)
            lastEma = 0
            lastTime = 0

            for i in range(arrlen):
                ema[i] = v[i] + lastEma * np.exp(-decay_rate * (t[i] - lastTime))
                lastEma = ema[i]
                lastTime = t[i]
            return ema

        # this is when two inputs specified via a tuple and...
        # this is when the output is preallocated via a dtype dictionary
        def get_ema_userfunc3(v, t, ema, decay_rate):
            lastEma = 0
            lastTime = 0
            for i in range(len(v)):
                ema[i] = v[i] + lastEma * np.exp(-decay_rate * (t[i] - lastTime))
                lastEma = ema[i]
                lastTime = t[i]

        decay = np.log(2) / (1e3 * 100)

        ema_result = ds.gb("sym").apply(get_ema_userfunc, decay_rate=decay)
        self.assertTrue(
            sum(correct - ema_result.ema) < 0.01,
            msg=f"got ema {ema_result['ema']}, wanted {correct}\nresult {sum(correct - ema_result.ema)}",
        )

        # here we use a tuple to pass in two values
        ema_result = ds.gb("sym").apply_nonreduce(get_ema_userfunc2, (ds.org,), decay_rate=decay)
        self.assertTrue(sum(correct - ema_result[0]) < 0.01)

        # here we use dtype with a dict to specify one output
        ema_result = ds.gb("sym").apply_nonreduce(
            get_ema_userfunc3, (ds.org,), decay_rate=decay, dtype={"ema": np.float64}
        )
        self.assertTrue(sum(correct - ema_result[0]) < 0.01)

        # we pass in tuple of a Dataset to specify the inputs
        # here we use dtype with a dict to specify one output
        ema_result = ds.sym.apply_nonreduce(
            get_ema_userfunc3,
            (ds[["org", "time"]],),
            decay_rate=decay,
            dtype={"ema": np.float64},
        )
        self.assertTrue(sum(correct - ema_result[0]) < 0.01)

        # we pass in tuple of two values
        # here we use dtype with a dict to specify one output
        # we change the output type to np.float32 and check the column name is 'ema'
        ema_result = ds.sym.apply_nonreduce(
            get_ema_userfunc3,
            (ds.org, ds.time),
            decay_rate=decay,
            dtype={"ema": np.float32},
        )
        self.assertTrue(
            sum(correct - ema_result["ema"]) < 0.01,
            msg=f"got ema {ema_result['ema']}, wanted {correct}",
        )

        # create another org
        ds.org2 = ds.org + 1
        ema_result = ds.gb("sym")["org", "org2"].apply_nonreduce(get_ema_userfunc2, (ds.time,), decay_rate=decay)
        self.assertTrue(sum(correct - ema_result["org"]) < 0.01)

        ema_result = ds.sym.apply_nonreduce(get_ema_userfunc2, [ds.org, ds.org2], (ds.time,), decay_rate=decay)
        self.assertTrue(sum(correct - ema_result["org"]) < 0.01)

    def test_apply_tuple(self):
        arrsize = 200
        numrows = 7

        ds = Dataset({"time": arange(arrsize * 1.0)})
        ds.data = arange(arrsize) % numrows
        ds.data2 = (arange(arrsize) + 3) % numrows
        symbols = [
            "AAPL",
            "AMZN",
            "FB",
            "GOOG",
            "IBM",
            "6",
            "7",
            "8",
            "9",
            "10",
            "11",
            "12",
            "13",
            "14",
            "15",
            "16",
            "17",
            "18",
        ]
        ds.symbol = Cat(1 + arange(arrsize) % len(symbols), symbols)

        # put the second argument in tuples
        result = ds.symbol.apply_reduce(lambda x, y: np.sum(np.minimum(x, y)), (ds.data, ds.data))[0]

        self.assertTrue(arr_eq(result[:4], np.asarray([33, 38, 32, 29])))

        # now test nanmin on integers
        ds.data[5] = np.nan
        ds.datau = ds.data.astype(np.uint64)
        result = ds.symbol.nansum(ds.data)[0]
        self.assertTrue(sum((result >= 0) & (result <= 100)) == 18)
        result = ds.symbol.nanmin(ds.data)[0]
        self.assertTrue(sum(result) == 0)
        result = ds.symbol.nanmax(ds.data)[0]
        self.assertTrue(sum(result == 6) == 18)
        result = ds.symbol.nansum(ds.datau)[0]
        self.assertTrue(sum((result >= 0) & (result <= 100)) == 18)
        result = ds.symbol.nanmin(ds.datau)[0]
        self.assertTrue(sum(result) == 0)
        result = ds.symbol.nanmax(ds.datau)[0]
        self.assertTrue(sum(result == 6) == 18)

        # test passing transform to apply_reduce
        ds.gb("data").apply_reduce(sum, transform=True)

    def test_apply_gb_order(self):
        ds = Dataset(
            {
                "Type": ["Put", "Call", "Put", "Put", "Call", "Call"],
                "Delta": [-0.7, 0.2, -0.4, -0.6, 0.3, 0.8],
                "PnL": [51, 3, 60, 529, 2, 36],
            }
        )
        maxpnl = ds.groupby(["Type"])["PnL"].max(transform=True)[0]
        self.assertTrue(maxpnl[0] == 529)
        self.assertTrue(maxpnl[1] == 36)

    def test_apply_gb_vs_cat(self):
        N = 100
        np.random.seed(1)
        ds = Dataset(
            dict(
                Symbol=Cat(np.random.choice(["SPY", "IBM"], N)),
                Exchange=Cat(np.random.choice(["AMEX", "NYSE"], N)),
                TradeSize=np.random.choice([1, 5, 10], N),
                TradePrice=np.random.choice([1.1, 2.2, 3.3], N),
            )
        )

        rds = ds.groupby(["Symbol", "Exchange"]).apply(
            lambda x: {
                "Count": len(x.TradeSize),
                "dollars_exec": (x.TradeSize * x.TradePrice).nansum(),
            }
        )

        c = ds.cat(["Symbol", "Exchange"], ordered=True)
        newds = c.count()
        newds["dollars_exec"] = c.apply_reduce(lambda x, y: (x * y).nansum(), ds.TradeSize, (ds.TradePrice,))
        newds
        self.assertTrue(newds.equals(rds))

        newds = ds.cat(["Symbol", "Exchange"], ordered=True).apply(
            lambda x: {
                "Count": len(x.TradeSize),
                "dollars_exec": (x.TradeSize * x.TradePrice).nansum(),
            },
            ds,
        )
        self.assertTrue(newds.equals(rds))

        r1 = ds.groupby(["Symbol", "Exchange"])["TradeSize"].apply_reduce(lambda x: len(x))[0]
        r2 = ds.groupby(["Symbol", "Exchange"]).count()[0]
        self.assertTrue(arr_eq(r1, r2))

    def test_apply_computable(self):
        ds = Dataset({"a": Cat(list("AABB")), "b": ["foo", "bar", "baz", "bazz"]}, unicode=True)
        result = ds["a"].apply_nonreduce(FA.shift, ds["b"], computable=False)
        self.assertTrue(result["b"][1] == "foo")

        ds.c = arange(4)
        ds["d"] = ds.a.copy()
        ds.c = arange(4)
        result = ds["a"].apply_nonreduce(FA.shift, ds[["b", "c", "d"]], computable=False)
        self.assertTrue(result["d"][1] == "A")

    def test_apply_dups(self):
        ds = Dataset({"a": FA(["B", "B"]), "b": FA(["Bb", "Bb"]), "i": FA([1, 1])})
        c = Cat([ds.a, ds.b])
        result = c.apply(np.sum, ds.i)
        self.assertEqual(result._ncols, 3)

        ds = Dataset({"a": FA(["B", "D"]), "b": FA(["Bb", "Bb"]), "i": FA([1, 1])})
        c = Cat([ds.a, ds.b])
        result = c.apply(np.sum, ds.i)
        self.assertEqual(result._ncols, 4)


if __name__ == "__main__":
    tester = unittest.main()



import sys
import unittest
from io import StringIO

import numpy as np

import riptable as rt


class Meta_Test(unittest.TestCase):
    def test_meta(self):
        st = rt.Struct(
            {
                "a": rt.Dataset(
                    {
                        "col1": rt.FastArray([1, 2]).astype(np.int32),
                        "col2": rt.FastArray([3, 4]).astype(np.int32),
                        "col4": rt.FastArray([5, 6]).astype(np.int32),
                    }
                ),
                "b": rt.FastArray([3, 4]).astype(np.int32),
            }
        )
        out = StringIO()
        orig_stdout = sys.stdout
        sys.stdout = out
        print(st.info())
        output = out.getvalue()
        target_output = """\x1b[1;36mDescription: \x1b[00m<no description>
\x1b[1;36mSteward: \x1b[00m<no steward>
\x1b[1;36mType: \x1b[00mStruct
\x1b[1;36mContents:\x1b[00m

\x1b[1;36mType     Name  Description                                         Steward     \x1b[00m
\x1b[1;36m-------  ----  --------------------------------------------------  ------------\x1b[00m
Dataset  \x1b[1;32ma   \x1b[00m  <no description>                                    <no steward>
int32    \x1b[1;32mb   \x1b[00m  <no description>                                    <no steward>
"""
        self.assertEqual(output, target_output)

        schema = {"Description": "This is a structure", "Steward": "Nick"}
        st.apply_schema(schema)
        st2 = rt.Struct({"This": st, "That": np.array([1, 2]).astype(np.int32)})
        out = StringIO()
        sys.stdout = out
        print(st2.info())
        output = out.getvalue()
        target_output = """\x1b[1;36mDescription: \x1b[00m<no description>
\x1b[1;36mSteward: \x1b[00m<no steward>
\x1b[1;36mType: \x1b[00mStruct
\x1b[1;36mContents:\x1b[00m

\x1b[1;36mType    Name  Description                                         Steward     \x1b[00m
\x1b[1;36m------  ----  --------------------------------------------------  ------------\x1b[00m
Struct  \x1b[1;32mThis\x1b[00m  This is a structure                                 Nick        
int32   \x1b[1;32mThat\x1b[00m  <no description>                                    <no steward>
"""
        self.assertEqual(output, target_output)

        schema = {
            "Description": "This is a structure",
            "Steward": "Nick",
            "Type": "Struct",
            "Contents": {
                "This": {
                    "Description": "This is a nested structure",
                    "Steward": "Bob",
                    "Type": "AttackHelicoptor",
                    "Contents": {
                        "a": {
                            "Description": "A description for a",
                            "Steward": "Fred",
                            "Contents": {
                                "col1": {
                                    "Description": "This describes column 1",
                                    "Steward": "Jay",
                                    "Type": "int32",
                                },
                                "col2": {
                                    "Description": "This describes column 2",
                                    "Steward": "Alex",
                                    "Type": "float32",
                                },
                                "col3": {
                                    "Description": "This column is not there",
                                    "Steward": "Ben",
                                },
                            },
                        },
                        "b": {
                            "Description": "A descriptiion for b",
                            "Steward": "George",
                        },
                    },
                },
                "That": {"Description": "This is an array", "Steward": "Willy"},
            },
        }

        res = st2.apply_schema(schema)
        res_c = {
            "This": {
                "Type Mismatch": "Type Struct does not match schema type AttackHelicoptor",
                "a": {
                    "col2": {"Type Mismatch": "Type int32 does not match schema type float32"},
                    "Extra Column": "col4",
                    "Missing Column": "col3",
                },
            }
        }
        self.assertEqual(res, res_c)

        out = StringIO()
        sys.stdout = out
        print(st2.info())
        output = out.getvalue()
        target_output = """\x1b[1;36mDescription: \x1b[00mThis is a structure
\x1b[1;36mSteward: \x1b[00mNick
\x1b[1;36mType: \x1b[00mStruct
\x1b[1;36mContents:\x1b[00m

\x1b[1;36mType    Name  Description                                         Steward     \x1b[00m
\x1b[1;36m------  ----  --------------------------------------------------  ------------\x1b[00m
Struct  \x1b[1;32mThis\x1b[00m  This is a nested structure                          Bob         
int32   \x1b[1;32mThat\x1b[00m  This is an array                                    Willy       
"""
        self.assertEqual(output, target_output)

        out = StringIO()
        sys.stdout = out
        print(st2.This.info())
        output = out.getvalue()
        target_output = """\x1b[1;36mDescription: \x1b[00mThis is a nested structure
\x1b[1;36mSteward: \x1b[00mBob
\x1b[1;36mType: \x1b[00mStruct
\x1b[1;36mContents:\x1b[00m

\x1b[1;36mType     Name  Description                                         Steward     \x1b[00m
\x1b[1;36m-------  ----  --------------------------------------------------  ------------\x1b[00m
Dataset  \x1b[1;32ma   \x1b[00m  A description for a                                 Fred        
int32    \x1b[1;32mb   \x1b[00m  A descriptiion for b                                George      
"""
        self.assertEqual(output, target_output)

        out = StringIO()
        sys.stdout = out
        print(st2.This.a.info())
        output = out.getvalue()
        target_output = """\x1b[1;36mDescription: \x1b[00mA description for a
\x1b[1;36mSteward: \x1b[00mFred
\x1b[1;36mType: \x1b[00mDataset
\x1b[1;36mContents:\x1b[00m

\x1b[1;36mType   Name  Description                                         Steward     \x1b[00m
\x1b[1;36m-----  ----  --------------------------------------------------  ------------\x1b[00m
int32  \x1b[1;32mcol1\x1b[00m  This describes column 1                             Jay         
int32  \x1b[1;32mcol2\x1b[00m  This describes column 2                             Alex        
int32  \x1b[1;32mcol4\x1b[00m  <no description>                                    <no steward>
"""
        self.assertEqual(output, target_output)

        out = StringIO()
        sys.stdout = out
        print(st2.This.a.col1.info())
        output = out.getvalue()
        target_output = """\x1b[1;36mDescription: \x1b[00mThis describes column 1
\x1b[1;36mSteward: \x1b[00mJay
\x1b[1;36mType: \x1b[00mint32
"""
        self.assertEqual(output, target_output)

        sys.stdout = orig_stdout


if __name__ == "__main__":
    tester = unittest.main()



__all__ = ["Item", "Info", "Doc", "apply_schema", "info", "doc"]

from typing import List, Optional

from .rt_display import DisplayText
from .rt_fastarray import FastArray
from .rt_struct import Struct

META_DICT = "_meta"
DOC_KEY = "Doc"
DESCRIPTION_KEY = "Description"
STEWARD_KEY = "Steward"
TYPE_KEY = "Type"
DETAIL_KEY = "Detail"
CONTENTS_KEY = "Contents"

NO_DESCRIPTION = "<no description>"
NO_STEWARD = "<no steward>"
NO_TYPE = "<none>"

NAME_DEFAULT_WIDTH = 4
DESCRIPTION_DEFAULT_WIDTH = 50
STEWARD_DEFAULT_WIDTH = 12
TYPE_STR_DEFAULT_WIDTH = 4

# ERROR KEYS
TYPE_MISMATCH = "Type Mismatch"
EXTRA_COLUMN = "Extra Column"
MISSING_COLUMN = "Missing Column"


class Item:
    """Descriptive information for a data object.

    Parameters
    ----------
    name : str
        The name of the data object.
    type : str
        The type of the data object.
    description : str
        A description of the data object.
    steward : str
        The steward of the data object.
    """

    name: str
    """str: The name of the data object."""
    type: str
    """str: The type of the data object."""
    description: str
    """str: A description of the data object."""
    steward: str
    """steward: The steward of the data object."""

    def __init__(self, name: str, type: str, description: str, steward: str):
        self.name = name
        self.type = type
        self.description = description
        self.steward = steward


class Info:
    """A hierarchically structured container of descriptive information
    for a data object.
    """

    title = []
    """list: The title of the data object"""
    description: Optional[str] = None
    """str: The description of the data object."""
    steward: Optional[str] = None
    """str: The steward of the data object."""
    type: Optional[str] = None
    """str: The type of the data object."""
    detail = None
    """str: Detail about the data object."""
    items: Optional[List[Item]] = None
    """list of `Item`: For a :class:`~.rt_struct.Struct` or :class:`~.rt_dataset.Dataset`, the items contained within it."""

    def __init__(self):
        pass

    def _make_text(self):
        title_format = DisplayText.title_format
        header_format = DisplayText.header_format

        rows = []
        if self.title:
            rows += [title_format("{}".format(self.title))]
            rows += [title_format("=" * len(self.title))]
        if self.description:
            rows += [header_format("Description: ") + self.description]
        if self.steward:
            rows += [header_format("Steward: ") + self.steward]
        if self.type:
            rows += [header_format("Type: ") + self.type]
        if self.detail:
            rows += [header_format("Detail: ") + self.detail]
        if self.items:
            rows += [header_format("Contents:"), ""]

            # Set column widths
            name_width = max(NAME_DEFAULT_WIDTH, max(len(item.name) for item in self.items))
            descrip_width = DESCRIPTION_DEFAULT_WIDTH
            steward_width = STEWARD_DEFAULT_WIDTH
            stype_width = max(TYPE_STR_DEFAULT_WIDTH, max(len(item.type) for item in self.items))

            # Add list header
            rows += [
                header_format(
                    "{: <{}}  {: <{}}  {: <{}}  {: <{}}".format(
                        "Type", stype_width, "Name", name_width, "Description", descrip_width, "Steward", steward_width
                    )
                )
            ]
            rows += [
                header_format(
                    "{}  {}  {}  {}".format(
                        "-" * stype_width, "-" * name_width, "-" * descrip_width, "-" * steward_width
                    )
                )
            ]

            # Add item rows
            for item in self.items:
                rows += [
                    "{: <{}}  {}  {: <{}}  {: <{}}".format(
                        item.type,
                        stype_width,
                        title_format("{: <{}}".format(item.name, name_width)),
                        item.description,
                        descrip_width,
                        item.steward,
                        steward_width,
                    )
                ]

        # Add a newline at the end if there is a title on top
        if self.title:
            rows += [""]

        return "\n".join(rows)

    def __str__(self):
        return DisplayText(self._make_text()).__str__()

    def __repr__(self):
        return DisplayText(self._make_text()).__repr__()

    def _repr_html_(self):
        return DisplayText(self._make_text())._repr_html_()


class Doc(Struct):
    """A document object containing metadata about a data object.

    Parameters
    ----------
    schema : dict
        See :meth:`apply_schema` for more information on the format of the
        dictionary.
    """

    _type = NO_TYPE
    _descrip = NO_DESCRIPTION
    _steward = NO_STEWARD
    _detail = None

    def __init__(self, schema):
        super().__init__()
        self._type = schema.get(TYPE_KEY)
        self._descrip = schema.get(DESCRIPTION_KEY, NO_DESCRIPTION)
        self._steward = schema.get(STEWARD_KEY, NO_STEWARD)
        self._detail = schema.get(DETAIL_KEY, None)
        schema_contents = schema.get(CONTENTS_KEY)
        if schema_contents:
            for key in schema_contents.keys():
                if self.is_valid_colname(key):
                    self[key] = Doc(schema_contents[key])

    def _as_info(self):
        info = Info()
        info.title = None
        info.description = self._descrip
        info.steward = self._steward
        info.type = self._type
        info.detail = self._detail
        info.items = []
        for name in self.keys():
            elem = self[name]
            info.items.append(Item(name, elem._type, elem._descrip, elem._steward))
        return info

    def __str__(self):
        return self._as_info().__str__()

    def __repr__(self):
        return self._as_info().__repr__()

    def _repr_html_(self):
        return self._as_info()._repr_html_()


def apply_schema(obj, schema: dict, doc: bool = True):
    """
    Apply a schema containing descriptive information recursively to the
    input data object.

    The schema should be in the form of a hierarchical dictionary, where
    for the data object, and recursively for each element it may contain,
    there is a descriptive dictionary with the following keys and values:

    * Type: 'Struct', 'Dataset', 'Multiset', 'FastArray', etc.
    * Description: a brief description of the data object
    * Steward: the name of the steward for that data object
    * Detail: any additional descriptive information
    * Contents: if the data object is a :class:`~.rt_struct.Struct`,
      :class:`~.rt_dataset.Dataset`, or :class:`~.rt_multiset.Multiset`, a
      recursively formed dictionary where there is a descriptive
      dictionary of this form associated with the name of each element
      contained by the data object.

    When the schema is applied to the data object, key/value pairs are set
    within the ``_meta`` dictionary attribute of the object and all of
    its elements, to enable subsequent retrieval of the descriptive
    information using the :meth:`.rt_struct.Struct.info` method or
    :meth:`.rt_struct.Struct.doc` property.

    In addition, during the schema application process, the contents and type
    of each data object is compared to the expectation of the schema, with
    any differences returned in the form of a dictionary.

    Parameters
    ----------
    obj : Struct or FastArray
        The data object to apply the schema information to.
    schema : dict
        A descriptive dictionary defining the schema that should apply to the
        data object and any elements it may contain.
    doc : bool
        Indicates whether to create and attach a :class:`Doc` to the object,
        so that the :meth:`doc` method may be run on the object.

    Returns
    -------
    res : dict
        Dictionary of deviations from the schema

    See Also
    --------
    :meth:`.rt_struct.Struct.apply_schema`
    """
    res = {}
    if isinstance(obj, (Struct, FastArray)):
        if not hasattr(obj, META_DICT):
            obj._meta = {}
        if doc:
            obj._meta[DOC_KEY] = Doc(schema)
        obj._meta[DESCRIPTION_KEY] = schema.get(DESCRIPTION_KEY, NO_DESCRIPTION)
        obj._meta[STEWARD_KEY] = schema.get(STEWARD_KEY, NO_STEWARD)
        obj._meta[DETAIL_KEY] = schema.get(DETAIL_KEY, None)
        stype = schema.get(TYPE_KEY)
        if stype and _type_str(obj) != stype:
            res[TYPE_MISMATCH] = "Type {} does not match schema type {}".format(_type_str(obj), stype)
        schema_contents = schema.get(CONTENTS_KEY)
        if schema_contents:
            for key in obj.keys():
                elem_schema = schema_contents.get(key)
                if elem_schema:
                    elem_res = apply_schema(obj[key], elem_schema, False)
                    if elem_res:
                        res[key] = elem_res
                else:
                    res[EXTRA_COLUMN] = key
            for key in schema_contents.keys():
                if key not in obj.keys():
                    res[MISSING_COLUMN] = key
    return res


def _type_str(obj) -> str:
    """
    Return the string representation of an object's type.

    Parameters
    ----------
    obj : Any
        An object

    Returns
    -------
    str : str
        String representation of an object's type.
    """
    if isinstance(obj, FastArray):
        stype = obj.dtype.name
    else:
        stype = type(obj).__name__
    return stype


def info(obj, title=None) -> Info:
    """
    Return the :class:`Info` for the object, describing its contents.

    Parameters
    ----------
    obj : Any
        The object
    title : str
        The title to give the object, defaults to None.

    Returns
    -------
    info : Info
        Information about `obj`.
    """
    info = Info()
    info.title = title
    info.description = NO_DESCRIPTION
    info.steward = NO_STEWARD
    info.detail = None
    info.type = _type_str(obj)
    if hasattr(obj, META_DICT):
        info.description = obj._meta.get(DESCRIPTION_KEY, info.description)
        info.steward = obj._meta.get(STEWARD_KEY, info.steward)
        info.detail = obj._meta.get(DETAIL_KEY, None)
    if isinstance(obj, Struct):
        info.items = []
        for name in obj.keys():
            descrip = NO_DESCRIPTION
            steward = NO_STEWARD
            if hasattr(obj[name], META_DICT):
                descrip = obj[name]._meta.get(DESCRIPTION_KEY, descrip)
                steward = obj[name]._meta.get(STEWARD_KEY, steward)
            info.items.append(Item(name, _type_str(obj[name]), descrip, steward))
    return info


def doc(obj) -> Optional[Doc]:
    """
    Return the :class:`Doc` for the object, describing its contents.

    Parameters
    ----------
    obj : Any
        The object.

    Returns
    -------
    doc : Doc
        Returns a :class:`Doc` instance if  the object contains documentation
        metadata, otherwise None.
    """
    if hasattr(obj, META_DICT):
        if DOC_KEY in obj._meta:
            return obj._meta[DOC_KEY]
    return None



__all__ = [
    "DATETIME_TYPES",
    "DS_DISPLAY_TYPES",
    "GB_FUNCTIONS",
    "MATH_OPERATION",
    "INVALID_DICT",
    "TIMEWINDOW_FUNCTIONS",
    "NumpyCharTypes",
    "REDUCE_FUNCTIONS",
    "ROLLING_FUNCTIONS",
    "SD_TYPES",
    "SM_DTYPES",
    "TypeRegister",
    "DisplayJustification",
    "DisplayColumnColors",
    "DisplayArrayTypes",
    "DisplayDetectModes",
    "DisplayLength",
    "ColHeader",
]


import sys
from collections import namedtuple
from enum import IntEnum
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
)

import numpy as np

if TYPE_CHECKING:
    from .rt_accum2 import Accum2
    from .rt_categorical import Categorical
    from .rt_dataset import Dataset
    from .rt_datetime import (
        Date,
        DateBase,
        DateSpan,
        DateTimeBase,
        DateTimeNano,
        TimeSpan,
    )
    from .rt_display import (
        DisplayAttributes,
        DisplayDetect,
        DisplayOptions,
        DisplayString,
        DisplayTable,
        DisplayText,
    )
    from .rt_fastarray import FastArray
    from .rt_groupby import GroupBy
    from .rt_grouping import Grouping
    from .rt_ledger import MathLedger
    from .rt_multiset import Multiset
    from .rt_pdataset import PDataset
    from .rt_sharedmemory import SharedMemory
    from .rt_sort_cache import SortCache
    from .rt_struct import Struct
    from .rt_timezone import Calendar, TimeZone


# common strings used to indicate special columns or items
INVALID_SHORT_NAME: str = "Inv"
INVALID_LONG_NAME: str = "Invalid"
TOTAL_LONG_NAME: str = "Total"
CLIPPED_LONG_NAME: str = "Clipped"
FILTERED_LONG_NAME: str = "Filtered"
GROUPBY_KEY_PREFIX: str = "key"
"""Default groupby key name - followed by _n for n"""

INVALID_POINTER_32: int = -2147483648
INVALID_POINTER_U32: int = 0xFFFFFFFF  # 4294967295
# INVALID_POINTER_64  = 0x8000000000000000  # -9223372036854775808
INVALID_POINTER_64: int = -9223372036854775808
INVALID_POINTER_U64: int = 0xFFFFFFFFFFFFFFFF  # 18446744073709551615


INVALID_DICT: Mapping[int, Any] = {
    # keys in this dict can be generated with ndarray.dtype.num
    0: False,  # np.bool
    1: -128,  # np.int8
    2: 255,  # numpy.uint8,
    3: -32768,  # numpy.int16
    4: 65535,  # numpy.uint16
    5: INVALID_POINTER_32,  # numpy.int32
    6: INVALID_POINTER_U32,  # numpy.uint32
    7: INVALID_POINTER_32,  # numpy.int_  # default for arange on linux and windows
    8: INVALID_POINTER_U32,  # numpy.uint_
    9: INVALID_POINTER_64,  # numpy.int64
    10: INVALID_POINTER_U64,  # np.uint64
    11: np.nan,  # np.float32
    12: np.nan,  # np.float64
    13: np.nan,  # np.float64
    # 14: numpy.complex64,   # np.complex64
    # 15: numpy.complex128,  # np.complex128
    # 16: numpy.complex128,  # np.complex128
    17: None,  # np.object_
    18: b"",  # np.bytes_
    19: "",  # np.str_
    # 20:                    # numpy.void
    # 21:                    # numpy.datetime64
    # 22:                    # numpy.timedelta64
    23: np.nan,  # numpy.float16
}

if sys.platform != "win32":
    INVALID_DICT: Mapping[int, Any] = {
        # keys in this dict can be generated with ndarray.dtype.num
        0: False,  # np.bool
        1: -128,  # np.int8
        2: 255,  # numpy.uint8,
        3: -32768,  # numpy.int16
        4: 65535,  # numpy.uint16
        5: INVALID_POINTER_32,  # numpy.int32
        6: INVALID_POINTER_U32,  # numpy.uint32
        7: INVALID_POINTER_64,  # numpy.int_  # default for arange on linux and windows
        8: INVALID_POINTER_U64,  # numpy.uint_
        9: INVALID_POINTER_64,  # numpy.int64
        10: INVALID_POINTER_U64,  # np.uint64
        11: np.nan,  # np.float32
        12: np.nan,  # np.float64
        13: np.nan,  # np.float64
        # 14: numpy.complex64,   # np.complex64
        # 15: numpy.complex128,  # np.complex128
        # 16: numpy.complex128,  # np.complex128
        17: None,  # np.object_
        18: b"",  # np.bytes_
        19: "",  # np.str_
        # 20:                    # numpy.void
        # 21:                    # numpy.datetime64
        # 22:                    # numpy.timedelta64
        23: np.nan,  # numpy.float16
    }


# See DatasetRW.h
# Official list of data types
class SM_DTYPES(IntEnum):
    DT_INVALID = 0
    DT_BOOL = 1
    DT_BYTE = 2
    DT_INT8 = 3
    DT_INT16 = 4
    DT_INT32 = 5
    DT_INT64 = 6
    DT_UINT8 = 7
    DT_UINT16 = 8
    DT_UINT32 = 9
    DT_UINT64 = 10
    DT_FLOAT16 = 11
    DT_FLOAT32 = 12
    DT_FLOAT64 = 13
    DT_OBJECT = 17
    DT_BYTES = 18  # fixed size ascii string uses this
    DT_UNICODE = 19
    DT_NPVOID = 20
    DT_DATETIME64 = 21
    DT_TIMEDELTA64 = 22
    DT_HALF = 23
    DT_CHARARRAY = 24  # how we like to store a fixed size ascii


class SD_TYPES(IntEnum):
    SD_UNKNOWN = 0

    # Matlab based
    SD_FUNCTIONH = 1
    SD_DATASET = 2
    SD_CLASS = 3
    SD_STRUCT = 4
    SD_SCALAR = 5
    SD_CHAR = 6
    SD_LOGICAL = 7
    SD_CELL = 8
    SD_NUMERIC = 9
    SD_VECTOR = 10

    # Python based
    SD_PANDAS = 20
    SD_NUMPY = 21


class MATH_OPERATION(IntEnum):
    """
    MATH_OPERATION is the encoding of the Riptable implemented mathematical operations.
    """

    # MATH_OPERATION is how Riptable communicates with RiptideCPP and this enumeration
    # is repeated in RiptideCPP CommonInc.h. Any changes made to either location must be
    # reflected in both places.
    # Two ops, returns same type
    ADD = 1
    SUB = 2
    MUL = 3
    MOD = 5
    MIN = 6
    MAX = 7
    NANMIN = 8
    NANMAX = 9
    FLOORDIV = 10
    POWER = 11
    REMAINDER = 12
    FMOD = 13

    # Two ops always return a float
    DIV = 101
    SUBDATETIMES = 102  # will check both sides for INV/ZEROS input:int64/int32 (returns double)
    # SUBDATETIMESRIGHT = 103,  # will check right hand side for ZEROS input:int64/int32 (returns double)
    SUBDATES = 103  # will check both sides for INV/ZEROS same dtype in/out (int32/int64)
    # ADDDATES = 105           # will check left hand side for ZEROS same dtype in/out (int32/int64)

    # One input returns same data type
    ABS = 201
    NEG = 202
    FABS = 203
    INVERT = 204
    FLOOR = 205
    CEIL = 206
    TRUNC = 207
    ROUND = 208
    NEGATIVE = 212
    POSITIVE = 213
    SIGN = 214
    RINT = 215
    EXP = 216
    EXP2 = 217
    # does not allow floats
    BITWISE_NOT = 218

    # One input always return a float one input
    SQRT = 301
    LOG = 302
    LOG2 = 303
    LOG10 = 304
    EXPM1 = 305
    LOG1P = 306
    SQUARE = 307
    CBRT = 308
    RECIPROCAL = 309

    # Two inputs Always return a bool
    CMP_EQ = 401
    CMP_NE = 402
    CMP_LT = 403
    CMP_GT = 404
    CMP_LTE = 405
    CMP_GTE = 406
    LOGICAL_AND = 407
    LOGICAL_XOR = 408
    LOGICAL_OR = 409

    # Two inputs
    BITWISE_LSHIFT = 501
    BITWISE_RSHIFT = 502
    BITWISE_AND = 503
    BITWISE_XOR = 504
    BITWISE_OR = 505
    BITWISE_ANDNOT = 506
    BITWISE_NOTAND = 507

    BITWISE_XOR_SPECIAL = 550

    # one input output bool
    LOGICAL_NOT = 601
    ISINF = 603
    ISNAN = 604
    ISFINITE = 605
    ISNORMAL = 606
    ISNOTINF = 607
    ISNOTNAN = 608
    ISNOTFINITE = 609
    ISNOTNORMAL = 610
    ISNANORZERO = 611
    SIGNBIT = 612


gBinaryUFuncs: Mapping[Callable, Optional[MATH_OPERATION]] = {
    # math ops
    np.add: MATH_OPERATION.ADD,
    np.subtract: MATH_OPERATION.SUB,
    np.multiply: MATH_OPERATION.MUL,
    np.matmul: None,
    np.divide: MATH_OPERATION.DIV,
    np.true_divide: MATH_OPERATION.DIV,
    np.floor_divide: MATH_OPERATION.FLOORDIV,
    np.remainder: MATH_OPERATION.REMAINDER,
    np.fmod: MATH_OPERATION.FMOD,
    np.mod: MATH_OPERATION.MOD,
    np.power: MATH_OPERATION.POWER,
    np.minimum: MATH_OPERATION.MIN,
    np.maximum: MATH_OPERATION.MAX,
    np.fmin: MATH_OPERATION.NANMIN,
    np.fmax: MATH_OPERATION.NANMAX,
}
"""
The mapping of Numpy to Riptable arithmetic binary operator overrides.

See Also
--------
MATH_OPERATION : the full set of mathematical operations supported by Riptable.
"""


gBinaryLogicalUFuncs: Mapping[Callable, MATH_OPERATION] = {
    # comparisons
    np.less_equal: MATH_OPERATION.CMP_LTE,
    np.less: MATH_OPERATION.CMP_LT,
    np.equal: MATH_OPERATION.CMP_EQ,
    np.not_equal: MATH_OPERATION.CMP_NE,
    np.greater: MATH_OPERATION.CMP_GT,
    np.greater_equal: MATH_OPERATION.CMP_GTE,
    np.logical_and: MATH_OPERATION.LOGICAL_AND,
    np.logical_xor: MATH_OPERATION.LOGICAL_XOR,
    np.logical_or: MATH_OPERATION.LOGICAL_OR,
}
"""
The mapping of Numpy to Riptable comparison function overrides.

See Also
--------
MATH_OPERATION : the full set of mathematical operations supported by Riptable.
"""


gBinaryBitwiseUFuncs: Mapping[Callable, Optional[MATH_OPERATION]] = {
    # bitwise operations only apply to bool and integers
    np.left_shift: None,
    np.right_shift: None,
    np.bitwise_and: MATH_OPERATION.BITWISE_AND,
    np.bitwise_xor: MATH_OPERATION.BITWISE_XOR,
    np.bitwise_or: MATH_OPERATION.BITWISE_OR,
}
"""
The mapping of Numpy to Riptable bit-twiddling binary operator overrides.

See Also
--------
MATH_OPERATION : the full set of mathematical operations supported by Riptable.
"""


gBinaryBitwiseMonoUFuncs: Mapping[Callable, Optional[MATH_OPERATION]] = {
    # bitwise operations only apply to bool and integers
    np.invert: None
}
"""
The mapping of Numpy to Riptable bit-twiddling unary operator overrides.

See Also
--------
MATH_OPERATION : the full set of mathematical operations supported by Riptable.
"""


gUnaryUFuncs: Mapping[Callable, Optional[MATH_OPERATION]] = {
    # math ops
    np.absolute: MATH_OPERATION.ABS,
    np.abs: MATH_OPERATION.ABS,
    np.fabs: MATH_OPERATION.FABS,
    np.invert: MATH_OPERATION.INVERT,
    np.floor: MATH_OPERATION.FLOOR,
    np.ceil: MATH_OPERATION.CEIL,
    np.trunc: MATH_OPERATION.TRUNC,
    np.round: MATH_OPERATION.ROUND,
    np.rint: MATH_OPERATION.ROUND,
    np.isinf: MATH_OPERATION.ISINF,
    np.isnan: MATH_OPERATION.ISNAN,
    np.isfinite: MATH_OPERATION.ISFINITE,
    np.signbit: MATH_OPERATION.SIGNBIT,
    np.negative: MATH_OPERATION.NEGATIVE,
    np.positive: MATH_OPERATION.POSITIVE,
    np.sign: MATH_OPERATION.SIGN,
    np.exp: MATH_OPERATION.EXP,
    np.exp2: MATH_OPERATION.EXP2,
    np.log: MATH_OPERATION.LOG,
    np.log2: MATH_OPERATION.LOG2,
    np.log10: MATH_OPERATION.LOG10,
    np.expm1: None,
    np.log1p: None,
    np.sqrt: MATH_OPERATION.SQRT,
    np.square: None,
    np.cbrt: MATH_OPERATION.CBRT,
    np.reciprocal: None,
    np.logical_not: MATH_OPERATION.LOGICAL_NOT,
    np.bitwise_not: MATH_OPERATION.BITWISE_NOT,
    np.signbit: MATH_OPERATION.SIGNBIT,
}
"""
The mapping of Numpy to Riptable arithmetic unary operator overrides.

See Also
--------
MATH_OPERATION : the full set of mathematical operations supported by Riptable.
"""


gNanFuncs: Mapping[Callable, Optional[Callable]] = {
    np.nanargmax: None,
    np.nanargmin: None,
    np.nancumprod: None,
    np.nancumsum: None,
    np.nanmax: None,
    np.nanmean: None,
    np.nanmedian: None,  # bug in numpy for floats?
    np.nanmin: None,
    np.nanpercentile: None,
    np.nanprod: None,
    np.nanstd: None,
    np.nansum: None,
    np.nanvar: None,
}


# See GroupBy.h
# Official list of funcss
class ROLLING_FUNCTIONS(IntEnum):
    ROLLING_SUM = 0
    ROLLING_NANSUM = 1
    ROLLING_MEAN = 102
    ROLLING_NANMEAN = 103
    ROLLING_QUANTILE = 104

    ROLLING_VAR = 106
    ROLLING_NANVAR = 107
    ROLLING_STD = 108
    ROLLING_NANSTD = 109


class REDUCE_FUNCTIONS(IntEnum):
    REDUCE_SUM = 0
    REDUCE_NANSUM = 1
    REDUCE_MEAN = 102
    REDUCE_NANMEAN = 103
    REDUCE_VAR = 106
    REDUCE_NANVAR = 107
    REDUCE_STD = 108
    REDUCE_NANSTD = 109
    REDUCE_MIN = 200
    REDUCE_NANMIN = 201
    REDUCE_MAX = 202
    REDUCE_NANMAX = 203
    REDUCE_ARGMIN = 204
    REDUCE_NANARGMIN = 205
    REDUCE_ARGMAX = 206
    REDUCE_NANARGMAX = 207

    # for Jack TODO
    REDUCE_ANY = 208
    REDUCE_ALL = 209


gReduceUFuncs: Mapping[Callable, REDUCE_FUNCTIONS] = {
    np.add: REDUCE_FUNCTIONS.REDUCE_SUM,
    np.minimum: REDUCE_FUNCTIONS.REDUCE_MIN,
    np.maximum: REDUCE_FUNCTIONS.REDUCE_MAX
    ## TODO add support for the following reduce ufuncs
    #   np.logical_and: RIPTIDE NP.ALL
    #   np.logical_or:  RIPTIDE NP.ANY
}
"""
The mapping of Numpy to Riptable reduce function overrides.

See Also
--------
REDUCE_FUNCTIONS encoding of all reduce functions that riptable supports
"""


class TIMEWINDOW_FUNCTIONS(IntEnum):
    TIMEWINDOW_SUM = 0
    TIMEWINDOW_PROD = 1


# See GroupBy.h
# Official list of funcss
class GB_FUNCTIONS(IntEnum):
    GB_SUM = 0
    GB_MEAN = 1
    GB_MIN = 2
    GB_MAX = 3

    # STD uses VAR with the param set to 1
    GB_VAR = 4
    GB_STD = 5

    GB_NANSUM = 50
    GB_NANMEAN = 51
    GB_NANMIN = 52
    GB_NANMAX = 53
    GB_NANVAR = 54
    GB_NANSTD = 55

    GB_FIRST = 100
    GB_NTH = 101
    GB_LAST = 102
    GB_MEDIAN = 103  # auto handles nan
    GB_MODE = 104  # auto handles nan
    GB_TRIMBR = 105  # auto handles nan
    GB_QUANTILE_MULT = 106  # handles all (nan)median/quantile versions

    # All int/uints output upgraded to INT64
    # Output is all elements (not just grouped)
    GB_ROLLING_SUM = 200
    GB_ROLLING_NANSUM = 201
    GB_ROLLING_DIFF = 202
    GB_ROLLING_SHIFT = 203
    GB_ROLLING_COUNT = 204
    GB_ROLLING_MEAN = 205
    GB_ROLLING_NANMEAN = 206
    GB_ROLLING_QUANTILE = 207

    # In ema.cpp
    GB_CUMSUM = 300
    GB_EMADECAY = 301
    GB_CUMPROD = 302
    GB_FINDNTH = 303
    GB_EMANORMAL = 304
    GB_EMAWEIGHTED = 305


# some groupby functions will work for strings
GB_STRING_ALLOWED = [
    GB_FUNCTIONS.GB_FIRST,
    GB_FUNCTIONS.GB_NTH,
    GB_FUNCTIONS.GB_LAST,
    GB_FUNCTIONS.GB_ROLLING_SHIFT,
    GB_FUNCTIONS.GB_MODE,
]
GB_DATE_ALLOWED = [
    GB_FUNCTIONS.GB_FIRST,
    GB_FUNCTIONS.GB_NTH,
    GB_FUNCTIONS.GB_LAST,
    GB_FUNCTIONS.GB_MEAN,
    GB_FUNCTIONS.GB_MIN,
    GB_FUNCTIONS.GB_MAX,
    GB_FUNCTIONS.GB_NANMEAN,
    GB_FUNCTIONS.GB_NANMIN,
    GB_FUNCTIONS.GB_NANMAX,
    GB_FUNCTIONS.GB_MEDIAN,
    GB_FUNCTIONS.GB_QUANTILE_MULT,
    GB_FUNCTIONS.GB_MODE,
    GB_FUNCTIONS.GB_ROLLING_DIFF,
    GB_FUNCTIONS.GB_ROLLING_SHIFT,
]

GB_FUNC_COUNT = -1
GB_FUNC_USER = 900
GB_FUNC_NUMBA = 1000


######################################################
# Numba groupby enums
######################################################
class GB_PACKUNPACK(IntEnum):
    UNPACK = 0
    PACK = 1


class NumpyCharTypes:
    All = "?bhilqpBHILQPefdgFDGSUVOMm"
    AllFloat = "efdgFDG"
    AllInteger = "bBhHiIlLqQpP"
    Computable = "fdgbBhHiIlLqQpP"  # does not include boolean or strings
    Noncomputable = "SeFDGUVOMm"
    Unsupported = "eFDGVOMm"  # unsupported in riptable world
    Supported = "?fdgbBhHiIlLqQpPSUV"
    SupportedFloat = "fdg"
    SupportedAlternate = "?fdgbBhHiIlLqQpPSU"
    Character = "c"
    Complex = "FDG"
    Datetime = "Mm"
    Float = "efdg"
    Float64 = "dg"
    Integer = "bhilqp"
    UnsignedInteger = "BHILQP"
    UnsignedInteger64 = "QP"
    SignedInteger64 = "qp"

    # linux gcc compiler long is int64, msvc long is int32
    if sys.platform != "win32":
        UnsignedInteger64 = "LQP"
        SignedInteger64 = "lqp"


gScalarType: Tuple[type, ...] = (
    int,
    float,
    complex,
    bool,
    bytes,
    str,
    memoryview,
    np.bool_,
    np.int8,
    np.uint8,
    np.int16,
    np.uint16,
    np.int32,
    np.uint32,
    np.int32,
    np.uint32,
    np.int64,
    np.uint64,
    np.float16,
    np.float32,
    np.float64,
    np.complex64,
    np.complex128,
    np.object_,
    np.bytes_,
    np.str_,
    np.void,
    np.datetime64,
    np.timedelta64,
)


gNumpyScalarType: Tuple[type, ...] = (
    memoryview,
    np.bool_,
    np.int8,
    np.uint8,
    np.int16,
    np.uint16,
    np.int32,
    np.uint32,
    np.int32,
    np.uint32,
    np.int64,
    np.uint64,
    np.float16,
    np.float32,
    np.float64,
    np.complex64,
    np.complex128,
    np.object_,
    np.bytes_,
    np.str_,
    np.void,
    np.datetime64,
    np.timedelta64,
)


def int_dtype_from_len(newlen: int) -> np.dtype:
    """Returns minimum itemsize unsigned integer dtype for given array length.
    Assumes that numbers up to the length will need to be stored by the returned dtype.
    Used by Grouping and Categorical.
    """
    if newlen < 100:
        dt = np.int8
    elif newlen < 30_000:
        dt = np.int16
    elif newlen < 2_000_000_000:
        dt = np.int32
    else:
        dt = np.int64
    return np.dtype(dt)


gAnsiColors: Mapping[str, str] = {
    "Black": "\x1b[0;30m",
    "BlinkBlack": "\x1b[5;30m",
    "BlinkBlue": "\x1b[5;34m",
    "BlinkCyan": "\x1b[5;36m",
    "BlinkGreen": "\x1b[5;32m",
    "BlinkLightGray": "\x1b[5;37m",
    "BlinkPurple": "\x1b[5;35m",
    "BlinkRed": "\x1b[5;31m",
    "BlinkYellow": "\x1b[5;33m",
    "Blue": "\x1b[0;34m",
    "Brown": "\x1b[0;33m",
    "Cyan": "\x1b[0;36m",
    "DarkGray": "\x1b[1;30m",
    "Green": "\x1b[0;32m",
    "LightBlue": "\x1b[1;34m",
    "LightCyan": "\x1b[1;36m",
    "LightGray": "\x1b[0;37m",
    "LightGreen": "\x1b[1;32m",
    "LightPurple": "\x1b[1;35m",
    "LightRed": "\x1b[1;31m",
    "NoColor": "",
    "Normal": "\x1b[0m",
    "Purple": "\x1b[0;35m",
    "Red": "\x1b[0;31m",
    "White": "\x1b[1;37m",
    "Yellow": "\x1b[1;33m",
}


class DS_DISPLAY_TYPES(IntEnum):
    HTML = 1
    REPR = 2
    STR = 3


class DATETIME_TYPES(IntEnum):
    ORDINAL_DATE = 1


DateTimeFormats: Mapping[str, str] = {"day": "%d-%b-%Y"}


class DisplayDetectModes(IntEnum):
    Jupyter = 1
    Ipython = 2
    Console = 3
    HTML = 5


class DisplayArrayTypes(IntEnum):
    Bool = 0
    Integer = 1
    Float = 2
    Bytes = 3
    Categorical = 4
    String = 5
    DateTime = 6
    DateTimeBase = 7
    DateTimeNano = 9
    TimeSpan = 10
    Record = 11


class DisplayLength(IntEnum):
    Undefined = 0
    Short = 1
    Medium = 2
    Long = 3


class TimeFormat(IntEnum):
    Clock = 1
    YearMonthDay = 2
    SIGNano = 3


class DisplayJustification(IntEnum):
    Undefined = 0
    Left = 1
    Right = 2
    Center = 3


class DisplayTextDecoration(IntEnum):
    Undefined = 0
    Bold = 1
    Italic = 2
    Underline = 3
    Strikethrough = 4


class DisplayNumberSeparator:
    NoSeparator = ""
    Comma = ","
    # Period      = "." #BUG, fix later
    Underscore = "_"


class DisplayColumnColors(IntEnum):
    Default = 0  # no styling
    Rownum = 1  # row numbers / default header color
    Sort = 2  # regular sort header and column data
    Groupby = 3  # groupby header and column data
    Multiset_head_a = 4  # comparison color for multiset columns headers
    Multiset_head_b = 5  # comparison color for multiset columns headers
    Multiset_col_a = 6  # comparison color for multiset column data
    Multiset_col_b = 7  # comparison color for multiset column data
    Accum2t = 8
    Purple = 9
    Pink = 10
    Red = 11
    GrayItalic = 12
    DarkBlue = 13
    BGColor = 14
    FGColor = 15


class ColumnStyle:
    """
    Holds display styles for entire columns or individual cells.
    These styles will override the defaults from the _display_query_properties callback in FastArray
    See also: DisplayColumn, DisplayCell, ItemFormat

    properties:
    color      : DisplayColumnColors
    align      : DisplayJustification
    decoration : DisplayTextDecoration
    width      : DisplayLength (default) OR can be set to new max width for array item's string repr

    """

    def __init__(
        self,
        color=DisplayColumnColors.Default,
        align=DisplayJustification.Right,
        decoration=DisplayTextDecoration.Undefined,
        width=None,
    ):
        self.color = color
        self.align = align
        self.decoration = decoration
        self.width = width

    def _build_string(self):
        repr_str = []
        repr_str.append(f"     {self.__class__.__name__}")
        repr_str.append(f"     color: {DisplayColumnColors(self.color).name}")
        repr_str.append(f"     align: {DisplayJustification(self.align).name}")
        repr_str.append(f"decoration: {DisplayTextDecoration(self.decoration).name}")
        repr_str.append(f"     width: {self.width}")
        return "\n".join(repr_str)

    def __repr__(self):
        return self._build_string()

    def __str__(self):
        return self._build_string()


class DisplayColorMode(IntEnum):
    NoColors = 0
    Light = 1
    Dark = 2


class CategoryMode(IntEnum):
    Default = 0
    StringArray = 1
    IntEnum = 2
    Dictionary = 3
    NumericArray = 4
    MultiKey = 5


class CategoricalOrigin(IntEnum):
    CategoricalView = 0
    CategoricalCopy = 1
    StringList = 2
    StringListWithCategories = 3
    NumericList = 4
    IndexWithCategories = 5
    CodeMapping = 6
    Multikey = 7
    Matlab = 8
    Pandas = 9
    SDSFile = 10


class CategoryStringMode(IntEnum):
    Default = 0
    Bytes = 1
    Unicode = 2


class CategoricalConstructor(IntEnum):
    EmptyValues = 0
    IntegerValues = 1
    FloatValues = 2
    StringValues = 3
    MultikeyListValues = 4
    MultikeyDictValues = 5


class ApplyType(IntEnum):
    Invalid = 0
    ReduceDataset = 1
    ReduceList = 2
    Dataset = 3
    Arrays = 4


class CompressionMode(IntEnum):
    Compress = 0
    Decompress = 1
    CompressFile = 2
    DecompressFile = 3
    SharedMemory = 4
    Info = 5


class CompressionType(IntEnum):
    Uncompressed = 0
    ZStd = 1


class ColumnAttribute(IntEnum):
    Default = 0
    Left = 1
    Right = 2


class SDSFlag(IntEnum):
    OriginalContainer = 0x01
    Stackable = 0x02
    Scalar = 0x04
    Nested = 0x08
    Meta = 0x10


class SDSFileType(IntEnum):
    Unknown = 0
    Struct = 1
    Dataset = 2
    Table = 3
    Array = 4
    OneFile = 5  # new for one file


class DayOfWeek(IntEnum):
    Monday = 0
    Tuesday = 1
    Wednesday = 2
    Thursday = 3
    Friday = 4
    Saturday = 5
    Sunday = 6


# allowing / now for denest
INVALID_FILE_CHARS: Tuple[str, ...] = ("\\", ":", "<", ">", "!", "|", "*", "?")


gBasicStats: Mapping[str, str] = {
    "count": "Number of non-null observations",
    "sum": "Sum of values",
    "mean": "Mean of values",
    "mad": "Mean absolute deviation",  # median(abs(a - median(a)))
    "median": "Arithmetic median of values",
    "min": "Minimum",
    "max": "Maximum",
    "std": "Unbiased standard deviation",
    "var": "Unbiased variance",
    "nansum": "Sum of values",
    "nanmean": "Mean of values",
    "nanmad": "Mean absolute deviation",
    "nanmedian": "Arithmetic median of values",
    "nanmin": "Minimum",
    "nanmax": "Maximum",
    "nanstd": "Unbiased standard deviation",
    "nanvar": "Unbiased variance",
    "mode": "Mode",
    "abs": "Absolute Value",
    "prod": "Product of values",
    "sem": "Unbiased standard error of the mean",
    "skew": "Unbiased skewness (3rd moment)",
    "kurt": "Unbiased kurtosis (4th moment)",
    "quantile": "Sample quantile (value at %)",
    "cumsum": "Cumulative sum",
    "cumprod": "Cumulative product",
    "cummax": "Cumulative maximum",
    "cummin": "Cumulative minimum",
}


#####################################################################################
# Structs begin
#####################################################################################


# Used in display table for multi-line column headers
# color_group indexing starts at 0
# cell_span cannot be 0.  a cell_span of 1 indicates 1 cell wide.
ColHeader = namedtuple("ColHeader", ["col_name", "cell_span", "color_group"])


###################################
## TJD NOTE: Need to use strings instead of enum here
##################################
class TypeId(IntEnum):
    Default = 0
    Struct = 1
    Dataset = 2
    Multiset = 3
    GroupBy = 4
    Grouping = 5
    FastArray = 6
    MathLedger = 7
    Categorical = 8
    Categories = 9
    Accum2 = 10
    DisplayDetect = 11
    DisplayOptions = 12
    DisplayTable = 13
    SortCache = 14
    DateTimeBase = 15
    DateTimeNano = 17
    TimeSpan = 18
    TimeZone = 19
    Calendar = 20
    Date = 21
    DateSpan = 22
    PDataset = 23


######################################################
# SDS File Header order
######################################################
gSDSFileHeader: List[str] = [
    "SDSHeaderMagic",
    "VersionHigh",
    "VersionLow",
    "CompMode",
    "CompType",
    "CompLevel",
    # ----- offset 16 -----
    "NameBlockSize",
    "NameBlockOffset",
    "NameBlockCount",
    "FileType",  # struct, dataset
    "AuthorId",  # python, matlab
    # ----- offset 48 -----
    "MetaBlockSize",
    "MetaBlockOffset",
    # ----- offset 64 -----
    "TotalMetaCompressedSize",
    "TotalMetaUncompressedSize",
    # ----- offset 80 -----
    "ArrayBlockSize",
    "ArrayBlockOffset",
    # ----- offset 96 -----
    "ArraysWritten",
    "ArrayFirstOffset",
    # ----- offset 112 -----
    "TotalArrayCompressedSize",
    "TotalArrayUncompressedSize",
]
SDS_EXTENSION: str = ".sds"
SDS_EXTENSION_BYTES: bytes = b".sds"


# please keep the TypeRegister at the end of the file
class TypeRegister:
    """
    When special classes are loaded, they register with this class to avoid cyclical dependencies
    """

    Struct: ClassVar[Type["Struct"]] = None
    Dataset: ClassVar[Type["Dataset"]] = None
    Multiset: ClassVar[Type["Multiset"]] = None
    GroupBy: ClassVar[Type["GroupBy"]] = None
    Grouping: ClassVar[Type["Grouping"]] = None
    FastArray: ClassVar[Type["FastArray"]] = None
    MathLedger: ClassVar[Type["MathLedger"]] = None
    Categorical: ClassVar[Type["Categorical"]] = None
    Categories: ClassVar[Type["Categories"]] = None
    Accum2: ClassVar[Type["Accum2"]] = None
    DisplayDetect: ClassVar[Type["DisplayDetect"]] = None
    DisplayOptions: ClassVar[Type["DisplayOptions"]] = None
    DisplayTable: ClassVar[Type["DisplayTable"]] = None
    DisplayString: ClassVar[Type["DisplayString"]] = None
    DisplayAttributes: ClassVar[Type["DisplayAttributes"]] = None
    DisplayText: ClassVar[Type["DisplayText"]] = None
    SortCache: ClassVar[Type["SortCache"]] = None
    DateTimeBase: ClassVar[Type["DateTimeBase"]] = None
    DateTimeNano: ClassVar[Type["DateTimeNano"]] = None
    TimeSpan: ClassVar[Type["TimeSpan"]] = None
    SharedMemory: ClassVar[Type["SharedMemory"]] = None
    TimeZone: ClassVar[Type["TimeZone"]] = None
    Calendar: ClassVar[Type["Calendar"]] = None
    DateBase: ClassVar[Type["DateBase"]] = None
    Date: ClassVar[Type["Date"]] = None
    DateSpan: ClassVar[Type["DateSpan"]] = None
    PDataset: ClassVar[Type["PDataset"]] = None

    @classmethod
    def validate_registry(cls):
        missing = set()
        for _nm in dir(cls):
            if not _nm.startswith("_") and getattr(TypeRegister, _nm) is None:
                missing.add(_nm)
        if len(missing) > 0:
            msg = ", ".join(sorted(missing))
            raise RuntimeError(f"riptable: Improper initialization!  Missing: {msg}")

    @classmethod
    def is_computable(cls, other):
        if not (other.dtype.char in NumpyCharTypes.Noncomputable or isinstance(other, cls.Categorical)):
            return True
        return False

    @classmethod
    def is_array_subclass(cls, arr):
        """
        Certain routines can be sped up by skipping the logic before falling back on a numpy call.
        Note: this is different than using python's issubclass(), which returns True if the classes are the same.
        Returns True if the item is an instance of a FastArray or numpy array subclass.
        """
        if isinstance(arr, np.ndarray):
            if type(arr) == np.ndarray or type(arr) == cls.FastArray:
                return False
            return True
        else:
            return False

    @classmethod
    def is_binned_array(cls, arr):
        """
        Use this instead of checking isinstance(item, TypeRegister.Categorical). For other binned
        types in the future.

        Called by:
        Dataset.melt()             -re-expands
        Dataset.from_jagged_rows() -re-expands
        GroupBy.__init__           -calls grouping, gb_keychain properties to borrow bins

        """
        return isinstance(arr, cls.Categorical)

    @classmethod
    def is_binned_type(cls, arrtype):
        """
        Check the type rather than the instance.
        See also is_binned_array()

        Called by:
        rt_utils._multistack_items()
        """
        return arrtype == cls.Categorical

    # ---------------------------------------------------------------
    @classmethod
    def is_spanlike(cls, arr: np.ndarray):
        """return True if it is a datespan or timespan"""
        # TODO: datetime/span are computable sometimes... need a way to distinguish from other FA subclasses
        # simple math works, but not larger groupby operations like sum
        result = False
        if isinstance(arr, (TypeRegister.TimeSpan, TypeRegister.DateSpan)):
            result = True
        return result

    # ---------------------------------------------------------------
    @classmethod
    def is_datelike(cls, arr: np.ndarray):
        """return True if it is a date or time"""
        # TODO: datetime/span are computable sometimes... need a way to distinguish from other FA subclasses
        # simple math works, but not larger groupby operations like sum
        result = False
        if isinstance(
            arr, (TypeRegister.DateTimeNano, TypeRegister.TimeSpan, TypeRegister.Date, TypeRegister.DateSpan)
        ):
            result = True
        return result

    @classmethod
    def is_string_or_object(cls, arr):
        return cls.is_array_subclass(arr) or arr.dtype.char in "OSU"

    @classmethod
    def newclassfrominstance(cls, instance, origin):
        """After slicing or an array routine, return a new instance of a FastArray subclass.
        If the array was not a subclass, instance is unchanged.

        Parameters
        ----------
        instance : ndarray
            Array generated from operation.
        origin : ndarray
            Array, possibly a FastArray subclass.

        Returns
        -------
        instance : ndarray
            Array of the same class as origin if the origin class has a newclassfrominstance defined.
        """
        # FastArray subclasses should define this classmethod to return
        # a new object with a different instance array
        if hasattr(origin, "newclassfrominstance"):
            instance = origin.newclassfrominstance(instance, origin)
        return instance

    # put these here for now
    # where should they live?
    @classmethod
    def as_meta_data(cls, obj):
        pass

    @classmethod
    def from_meta_data(cls, itemdict: Optional[dict] = None, flags: Optional[list] = None, meta: str = ""):
        from .Utils.rt_metadata import MetaData

        if itemdict is None:
            itemdict = dict()
        if flags is None:
            flags = list()
        meta = MetaData(meta)
        iclass = meta.itemclass
        return iclass._from_meta_data(itemdict, flags, meta)



import os
import unittest

import pytest

from riptable import *
from riptable.rt_sds import SDSMakeDirsOn

# change to true since we write into /tests directory
SDSMakeDirsOn()


class PDataset_Test(unittest.TestCase):
    # TODO use pytest file fixtures
    def test_constructor_ds_list(self):
        correct_cutoffs = [5, 10, 15]
        correct_auto = ["p0", "p1", "p2"]
        correct_custom = ["test1", "test2", "test3"]
        correct_files = ["f20180201.sds", "f20190201.sds", "f20200201.sds"]
        correct_dates = ["20180201", "20190201", "20200201"]

        ds = Dataset({"col_" + str(i): arange(5) for i in range(5)})

        # no info
        pds = PDataset([ds, ds, ds])
        self.assertTrue(pds._pfilenames == [])
        self.assertTrue(bool(np.all(pds.pnames == correct_auto)))
        self.assertTrue(bool(np.all(pds.pcutoffs == correct_cutoffs)))
        self.assertEqual(pds.pcount, 3)
        self.assertEqual(pds._nrows, 15)
        self.assertTrue(isinstance(pds._pnames, dict))

        # with names
        pds = PDataset([ds, ds, ds], pnames=["test1", "test2", "test3"])
        self.assertTrue(pds._pfilenames == [])
        self.assertTrue(bool(np.all(pds.pnames == correct_custom)))
        self.assertTrue(bool(np.all(pds.pcutoffs == correct_cutoffs)))
        self.assertEqual(pds.pcount, 3)
        self.assertEqual(pds._nrows, 15)
        self.assertTrue(isinstance(pds._pnames, dict))

        # with filenames that have dates
        pds = PDataset([ds, ds, ds], filenames=correct_files)
        self.assertTrue(pds._pfilenames == correct_files)
        self.assertTrue(bool(np.all(pds.pnames == correct_dates)))
        self.assertTrue(bool(np.all(pds.pcutoffs == correct_cutoffs)))
        self.assertEqual(pds.pcount, 3)
        self.assertEqual(pds._nrows, 15)
        self.assertTrue(isinstance(pds._pnames, dict))

        # with filenames AND pnames
        pds = PDataset([ds, ds, ds], pnames=correct_custom, filenames=correct_files)
        self.assertTrue(pds._pfilenames == correct_files)
        self.assertTrue(bool(np.all(pds.pnames == correct_custom)))
        self.assertTrue(bool(np.all(pds.pcutoffs == correct_cutoffs)))
        self.assertEqual(pds.pcount, 3)
        self.assertEqual(pds._nrows, 15)
        self.assertTrue(isinstance(pds._pnames, dict))

    def test_constructor_single(self):
        ds = Dataset({"col_" + str(i): arange(5) for i in range(5)})
        pds = PDataset(ds)
        self.assertTrue(pds._pfilenames == [])
        self.assertTrue(bool(np.all(pds.pnames == ["p0"])))
        self.assertEqual(pds.pcutoffs[0], 5)
        self.assertEqual(pds.pcount, 1)
        self.assertTrue(isinstance(pds._pnames, dict))

    def test_constructor_file_list(self):
        correct_cutoffs = [5, 10, 15]
        correct_auto = ["p0", "p1", "p2"]
        correct_custom = ["test1", "test2", "test3"]
        ds = Dataset({"col_" + str(i): arange(5) for i in range(5)})
        paths = [rb"riptable/tests/temp/ds" + str(i).encode() + b".sds" for i in range(3)]
        for i in range(3):
            ds.save(paths[i])

        # filepath only (found changing number)
        pds = PDataset(paths)
        self.assertTrue(pds._pfilenames == paths)
        self.assertTrue(
            bool(np.all(pds.pnames == ["0", "1", "2"])),
            msg=f"{pds.pnames} vs. {correct_auto}",
        )
        self.assertTrue(bool(np.all(pds.pcutoffs == correct_cutoffs)))
        self.assertEqual(pds.pcount, 3)
        self.assertEqual(pds._nrows, 15)
        self.assertTrue(isinstance(pds._pnames, dict))

        # filepath with custom pnames
        pds = PDataset(paths, pnames=correct_custom)
        self.assertTrue(pds._pfilenames == paths)
        self.assertTrue(
            bool(np.all(pds.pnames == correct_custom)),
            msg=f"{pds.pnames} vs. {correct_auto}",
        )
        self.assertTrue(bool(np.all(pds.pcutoffs == correct_cutoffs)))
        self.assertEqual(pds.pcount, 3)
        self.assertEqual(pds._nrows, 15)
        self.assertTrue(isinstance(pds._pnames, dict))

        for p in paths:
            os.remove(p)

    def test_files_no_dates(self):
        ds = Dataset({"col_" + str(i): arange(5) for i in range(5)})
        correct_auto = ["p0", "p1", "p2"]
        paths = [rb"riptable/tests/temp/ds" + str(i).encode() + b".sds" for i in ["a", "b", "c"]]
        for i in range(3):
            ds.save(paths[i])

        pds = PDataset(paths)
        self.assertTrue(bool(np.all(pds.pnames == correct_auto)))

        for p in paths:
            os.remove(p)

    @pytest.mark.skip(reason="this test no longer works and is disabled")
    def test_constructor_errors(self):
        garbage = [1, 2, "test.sds"]
        with self.assertRaises(TypeError):
            pds = PDataset(garbage)

        garbage = [1, 2, 3]
        with self.assertRaises(TypeError):
            pds = PDataset(garbage)

    # will add more extensive indexing tests later
    def test_copy_internal(self):
        ds = Dataset({"col_" + str(i): arange(5) for i in range(5)})
        pds = PDataset([ds, ds])

        # if row index, return dataset
        pds2 = pds[:2, :]
        self.assertTrue(type(pds2) == Dataset)

        # if col index, return pdataset
        pds2 = pds[:2]
        self.assertTrue(isinstance(pds2, PDataset))

    def test_prows(self):
        ds1 = Dataset({"col_" + str(i): arange(5) for i in range(5)})
        ds2 = Dataset({"col_" + str(i): arange(7) for i in range(5)})
        ds3 = Dataset({"col_" + str(i): arange(9) for i in range(5)})

        pds = PDataset([ds1, ds2, ds3])
        self.assertTrue(bool(np.all(pds.prows == [5, 7, 9])))

    def test_pdict(self):
        names = ["test1", "test2", "test3"]
        slices = [slice(0, 5, None), slice(5, 12, None), slice(12, 21, None)]
        ds1 = Dataset({"col_" + str(i): arange(5) for i in range(5)})
        ds2 = Dataset({"col_" + str(i): arange(7) for i in range(5)})
        ds3 = Dataset({"col_" + str(i): arange(9) for i in range(5)})

        pds = PDataset([ds1, ds2, ds3], pnames=names)
        for i, (k, v) in enumerate(pds.pdict.items()):
            self.assertEqual(k, names[i])
            self.assertTrue(v == slices[i])

    # disabled while building partitioned groupby
    # def rtest_pgb(self):
    #    ds1 = Dataset({'col_'+str(i):arange(5) for i in range(5)})
    #    ds2 = Dataset({'col_'+str(i):arange(7) for i in range(5)})
    #    ds3 = Dataset({'col_'+str(i):arange(9) for i in range(5)})

    #    pds = PDataset([ds1,ds2,ds3])
    #    g = pds.pgb
    #    self.assertTrue(isinstance(g,GroupBy))
    #    result = g.sum()
    #    self.assertTrue('Partition' in pds)
    #    self.assertTrue(isinstance(pds.Partition,Categorical))
    #    self.assertEqual(result.label_get_names()[0], 'Partition')

    def test_piter(self):
        ds1 = Dataset({"col_" + str(i): arange(5) for i in range(5)})
        ds2 = Dataset({"col_" + str(i): arange(7) for i in range(5)})
        ds3 = Dataset({"col_" + str(i): arange(9) for i in range(5)})

        dlist = [ds1, ds2, ds3]
        pds = PDataset(dlist)

        for i, (name, ds) in enumerate(pds.piter):
            self.assertEqual(name, pds.pnames[i])
            self.assertTrue(type(ds) == Dataset)
            eq = ds == dlist[i]
            for col in eq.values():
                self.assertTrue(bool(np.all(col)))

    def test_pslices(self):
        ds1 = Dataset({"col_" + str(i): arange(5) for i in range(5)})
        ds2 = Dataset({"col_" + str(i): arange(7) for i in range(5)})
        ds3 = Dataset({"col_" + str(i): arange(9) for i in range(5)})
        slices = [slice(0, 5, None), slice(5, 12, None), slice(12, 21, None)]

        dlist = [ds1, ds2, ds3]
        pds = PDataset(dlist)

        for i, s in enumerate(pds.pslices):
            self.assertEqual(s, slices[i])

    def test_pslice(self):
        ds1 = Dataset({"col_" + str(i): arange(5) for i in range(5)})
        ds2 = Dataset({"col_" + str(i): arange(7) for i in range(5)})
        ds3 = Dataset({"col_" + str(i): arange(9) for i in range(5)})
        slices = [slice(0, 5, None), slice(5, 12, None), slice(12, 21, None)]

        dlist = [ds1, ds2, ds3]
        pds = PDataset(dlist)

        for i in range(3):
            s = pds.pslice(i)
            self.assertEqual(slices[i], s)

        with self.assertRaises(IndexError):
            s = pds.pslice("a")

    def test_partition(self):
        ds1 = Dataset({"col_" + str(i): arange(5) for i in range(5)})
        ds2 = Dataset({"col_" + str(i): arange(7) for i in range(5)})
        ds3 = Dataset({"col_" + str(i): arange(9) for i in range(5)})

        dlist = [ds1, ds2, ds3]
        pds = PDataset(dlist)

        for i, name in enumerate(pds.pnames):
            ds = pds.partition(name)
            self.assertTrue(type(ds) == Dataset)
            eq = ds == dlist[i]
            for col in eq.values():
                self.assertTrue(bool(np.all(col)))

        for i in range(3):
            ds = pds.partition(i)
            self.assertTrue(type(ds) == Dataset)
            eq = ds == dlist[i]
            for col in eq.values():
                self.assertTrue(bool(np.all(col)))

        with self.assertRaises(IndexError):
            _ = pds.partition([])

    def test_except_index(self):
        ds1 = Dataset({"col_" + str(i): arange(5) for i in range(5)})
        pds = PDataset(ds1, pnames=["123"])

        ds2 = pds[123]
        eq = ds1 == ds2
        for col in eq.values():
            self.assertTrue(bool(np.all(col)))

        with self.assertRaises(KeyError):
            p = pds[1234]

        with self.assertRaises(KeyError):
            p = pds[[True, False]]

    def test_pload(self):
        ds1 = Dataset({"col_" + str(i): arange(5) for i in range(5)})
        ds2 = Dataset({"col_" + str(i): arange(7) for i in range(5)})
        ds3 = Dataset({"col_" + str(i): arange(9) for i in range(5)})
        dlist = [ds1, ds2, ds3]

        correct_files = ["f20180201.sds", "f20180202.sds", "f20180203.sds"]
        correct_dates = ["20180201", "20180202", "20180203"]
        path = "riptable/tests/temp/f{}.sds"

        for i, fname in enumerate(correct_files):
            dlist[i].save(r"riptable/tests/temp/" + fname)

        pds = PDataset.pload(path, 20180201, 20180203)
        self.assertEqual(pds.pcount, 3)
        self.assertTrue(bool(np.all(pds.pnames == correct_dates)))

        for i, fname in enumerate(correct_files):
            os.remove(r"riptable/tests/temp/" + fname)

    def test_psave(self):
        ds1 = Dataset({"col_" + str(i): arange(5) for i in range(5)})
        ds2 = Dataset({"col_" + str(i): arange(7) for i in range(5)})
        ds3 = Dataset({"col_" + str(i): arange(9) for i in range(5)})
        dlist = [ds1, ds2, ds3]
        pds = PDataset(dlist)
        with self.assertRaises(NotImplementedError):
            pds.psave()

    def test_partition_rename(self):

        ds = Dataset({"col_" + str(i): arange(5) for i in range(5)})
        pds = PDataset([ds, ds, ds])
        newnames = ["part1", "part2", "part3"]
        pds.set_pnames(newnames)

        self.assertTrue(bool(np.all(pds.pnames == newnames)))

        with self.assertRaises(ValueError):
            pds.set_pnames(newnames[:2])

        with self.assertRaises(ValueError):
            pds.set_pnames(["p0", "p0", "p0"])

        with self.assertRaises(ValueError):
            pds.set_pnames(dict(zip(newnames, [0, 0, 0])))


if __name__ == "__main__":
    tester = unittest.main()



import random as rand
import unittest

import pandas as pd

import riptable as rt

functions_str = [
    "count",
    "sum",
    "mean",
    "median",
    "min",
    "max",
    # 'prod',
    "var",
    "quantile",
    "cumsum",
    "cumprod",
    # 'cummax',
    # 'cummin'
    "first",
    "last",
    # 'mode'
]

# dictionary of {func_name : (rt_cat_kwargs, pd_gb_kwargs)}
functions_kwargs = {"quantile": ({"q": 0.84}, {"q": 0.84, "interpolation": "midpoint"})}

import numpy as np

type_list = [
    # np.bool,          ## not a numeric type
    np.intc,
    np.intp,
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
    # np.float16,       ## not supported
    np.float32,
    np.float64,
    # np.complex64,     ## not supported
    # np.complex128     ## not supported
]


def safe_equal(ary1, ary2):
    assert len(ary1) == len(ary2)

    def isNaN(num):
        return num != num

    for a, b in zip(ary1, ary2):
        if not (a == b or (isNaN(a) and isNaN(b))):
            return False
    return True


def min(a, b):
    return a if a < b else b


class GroupbyFunctions_Test(unittest.TestCase):
    def groupby_func(self, df, fn):
        return getattr(df, functions_str[fn])

    def test_single_col_groupby_tests(self):

        Values = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]
        Keys = ["a", "b", "c", "a", "b", "c", "d", "e", "f"]
        for type_ in type_list:

            data = {"Vs": rt.FastArray(Values, dtype=type_), "Ks": Keys}

            pd_data = pd.DataFrame(data)
            sfw_data = rt.Dataset(data)

            key = "Ks"
            val = "Vs"

            pd_gb = pd_data.groupby(key)
            sfw_gb = sfw_data.groupby(key)

            for name in functions_str:

                pd_func = getattr(pd_gb, name)
                sfw_func = getattr(sfw_gb, name)

                pd_kwargs = {}
                sfw_kwargs = {}

                if name in functions_kwargs:
                    sfw_kwargs, pd_kwargs = functions_kwargs[name]

                pd_out = pd_func(**pd_kwargs)
                sfw_out = sfw_func(**sfw_kwargs)

                pd_col = pd_out[val]._values
                if name == "count":
                    sfw_col = sfw_out["Count"]
                else:
                    sfw_col = sfw_out[val]

                is_integer_subttype = np.issubdtype(type_, np.integer)
                # median for integer was fixed, so removed previous check here
                if not safe_equal(pd_col, sfw_col):
                    print("data_type_t = ", type_)
                    print("function =", name)
                    print("pandas output =", pd_col)
                    print("sfw    output =", sfw_col)
                    # TODO move as error message following assert
                    self.assertTrue(False)

    # TODO pytest parameterize type_list
    def test_multi_col_groupby_tests(self, numb_keys_and_values=5, numb_rows=20):
        col_val_names = ["alpha", "beta", "gamma", "sigma", "zeta"]
        col_key_names = ["lions", "tigers", "bears", "oh", "my"]

        MAX_LENGTH = min(len(col_val_names), len(col_key_names))
        assert numb_keys_and_values <= MAX_LENGTH
        for type_ in type_list:

            vals = [1, 2, 3, 4, 5, 6, 7, 8, 9, 0]
            keys = "a b c d e f g".split(" ")

            vs = []
            ks = []

            for i in range(0, numb_keys_and_values):
                vs.append([vals[rand.randint(0, len(vals) - 1)] for i in range(0, numb_rows)])
                ks.append([keys[rand.randint(0, len(keys) - 1)] for i in range(0, numb_rows)])

            data = {}

            for i in range(0, numb_keys_and_values):
                data[col_val_names[i]] = rt.FastArray(vs[i], dtype=type_)
                data[col_key_names[i]] = rt.FastArray(vs[i], dtype=type_)

            pd_data = pd.DataFrame(data)
            sfw_data = rt.Dataset(data)
            key = col_key_names[0:numb_keys_and_values]
            val = col_val_names[0:numb_keys_and_values]

            pd_gb = pd_data.groupby(key)
            sfw_gb = sfw_data.groupby(key)

            for name in functions_str:
                pd_func = getattr(pd_gb, name)
                sfw_func = getattr(sfw_gb, name)

                pd_kwargs = {}
                sfw_kwargs = {}

                if name in functions_kwargs:
                    pd_kwargs, sfw_kwargs = functions_kwargs[name]

                pd_out = pd_func(**pd_kwargs)
                sfw_out = sfw_func(**sfw_kwargs)

                if name == "count":
                    # only compare one column for count
                    pd_col = pd_out["alpha"]
                    sfw_col = sfw_out.Count
                    if not safe_equal(pd_col, sfw_col):
                        print("function =", name)
                        print("pandas output =", pd_col)
                        print("sfw    output =", sfw_col)
                        self.assertTrue(False)
                else:
                    for val in col_val_names:
                        # extract array from pandas series
                        pd_col = pd_out[val]._values
                        sfw_col = sfw_out[val]

                        is_integer_subttype = np.issubdtype(type_, np.integer)
                        is_median = name != "median"
                        if not safe_equal(pd_col, sfw_col) and (not is_integer_subttype and not is_median):
                            print("function =", name)
                            print("pandas output =", pd_col)
                            assert False


if __name__ == "__main__":
    tester = unittest.main()
